import re


def uncapitalize(str):
    return str[0:1].lower() + str[1:]


def dbl_qutes_str(str):
    return f'"{str}"'


def qutes_str(str):
    return f"'{str}'"


def delete_dbl_spaces(value):
    if value == None:
        return value
    return re.sub('\s+', ' ', value).strip();


class Common:
    @staticmethod
    def get_size_file_str(value):
        if value == 0:
            return ""

        if value > 0 and value <= 1024:
            return f'{value} Байт'

        if value > 1024 and value <= 1024 * 1024:
            return f'{round(value / 1024, 2)} КБайт'

        if value > 1024 * 1024 and value <= 1024 * 1024 * 1024:
            return f'{round(value / 1024 / 1024, 2)} МБайт'

        if value > 1024 * 1024 * 1024:
            return f'{round(value / 1024 / 1024 / 1024, 2)}  ГБайт'
        return value

    @staticmethod
    def arraund_error(error):
        str = '!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!'
        return f'\n{str}\n{error}\n{str}'



import os

# import websockets

from isc_common.http.DSRequest import DSRequest


class DSResponse_UploadFile(DSRequest):
    def remove(self, path):
        if os.path.exists(path):
            os.remove(path)

    def get_path(self, name):
        return os.path.dirname(os.path.abspath(name)).replace(os.sep, os.altsep) if os.altsep else os.path.dirname(os.path.abspath(name))



from django.db.models import TextField

from isc_common import delAttr
from isc_common.fields import Field


class DescriptionField(TextField, Field):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('null', True)
        kwargs.setdefault('blank', True)
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        delAttr(kwargs, 'null')
        delAttr(kwargs, 'blank')
        return name, path, args, kwargs



import logging

from django.conf import settings
from django.core.management import BaseCommand
from django.db import connection
from transliterate import translit
from transliterate.exceptions import LanguageDetectionError

from one_c.models.entities import Entities, Field
from one_c.models.entity_1c import Entity_1c
from one_c.models.one_c_params_entity_view import One_c_params_entity_view

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    entities = settings.ONE_C_ENTITIES()

    def dbl_qutes_str(self, str):
        return f'"{str}"'

    def qutes_str(self, str):
        return f"'{str}'"

    def uncapitalize(self, str):
        return str[0:1].lower() + str[1:]

    def handle(self, *args, **options):

        try:
            fields = []
            sql = []
            # sql1 = []
            for entity in Entity_1c.objects.filter():
                entity_code = translit(entity.code, reversed=True).replace("'", '')

                m_view_name = f'{self.dbl_qutes_str(entity_code + "_mview")}'

                sql.append(f'DROP MATERIALIZED VIEW IF EXISTS {m_view_name};')
                sql.append(f'CREATE MATERIALIZED VIEW {m_view_name} AS SELECT d.ref, d.entity_id,')
                sql.append(f'$FIELDS')

                for param in One_c_params_entity_view.objects.filter(entity=entity):
                    used_field = self.entities.get_field(entity.id, param.code)

                    if used_field:
                        if used_field.type == Field.string():
                            item = f'(SELECT string_agg(dp.value, {self.qutes_str(", ")}) FROM ducument_param_1c_view dp JOIN one_c_param_type pt ON pt.id = dp.type_id WHERE dp.document_id = d.ref AND pt.code::text = ' \
                                f'{self.qutes_str(param.code)}::text) AS {self.dbl_qutes_str(self.uncapitalize(used_field.translited_alias))}'
                        elif used_field.type == Field.uuid():
                            if param.code != 'Ref':
                                item = f'(SELECT case when dp.value_uuid::text = {self.qutes_str("00000000-0000-0000-0000-000000000000")} then null else dp.value_uuid end as value_uuid FROM ducument_param_1c_view dp JOIN one_c_param_type pt ON pt.id = dp.type_id WHERE dp.document_id = d.ref AND pt.code::text = ' \
                                    f'{self.qutes_str(param.code)}::text) AS {self.dbl_qutes_str(self.uncapitalize(used_field.translited_alias))}'
                            else:
                                item = None
                        elif used_field.type == Field.int():
                            item = f'(SELECT dp.value_int FROM ducument_param_1c_view dp JOIN one_c_param_type pt ON pt.id = dp.type_id WHERE dp.document_id = d.ref AND pt.code::text = ' \
                                f'{self.qutes_str(param.code)}::text) AS {self.dbl_qutes_str(self.uncapitalize(used_field.translited_alias))}'
                        elif used_field.type == Field.float():
                            item = f'(SELECT dp.value_float FROM ducument_param_1c_view dp JOIN one_c_param_type pt ON pt.id = dp.type_id WHERE dp.document_id = d.ref AND pt.code::text = ' \
                                f'{self.qutes_str(param.code)}::text) AS {self.dbl_qutes_str(self.uncapitalize(used_field.translited_alias))}'
                        elif used_field.type == Field.boolean():
                            item = f'(SELECT dp.value_boolean FROM ducument_param_1c_view dp JOIN one_c_param_type pt ON pt.id = dp.type_id WHERE dp.document_id = d.ref AND pt.code::text = ' \
                                f'{self.qutes_str(param.code)}::text) AS {self.dbl_qutes_str(self.uncapitalize(used_field.translited_alias))}'
                        else:
                            raise Exception(f'Unknown type: {used_field.type}')

                        if item:
                            fields.append(item)

                sql.append(f'FROM one_c_document_1c d WHERE d.entity_id = {entity.id} WITH NO DATA;')
                sql.append(f'REFRESH MATERIALIZED VIEW {m_view_name};')

                sql_str = '\n'.join(sql).replace('$FIELDS', ',\n'.join(fields))

                with connection.cursor() as cursor:
                    logger.debug(f'\n{sql_str}')
                    cursor.execute(sql_str)
                    logger.debug(f'Created all')

        except Exception as ex:
            raise ex



from django.db import models

from code_generator.app import model
from code_generator.app.app_mailwindow import AppMainWindow


def os(args):
    pass


class App(model.Model):


    databases_templ = 'ts/datasources/datasource.pmd'
    views_templ = 'python/views.pmd'
    urls_templ = 'python/urls.pmd'

    databases = []
    editors = []
    views = []
    urls = []
    templates = []
    others = []

    def write_entity(self):
        for key, value in self.get_model_attributes():
            if isinstance(value, AppMainWindow):
                setattr(value, 'name_app', self.name_app)
                setattr(value, 'code_app', self.code_app)
                setattr(value, 'path_app', self.path_app)
                value.write_entity()

        for key, value in self.get_model_attributes():
            if isinstance(value, models.Model):
                pass



from django.db.models import CharField
from isc_common.fields import Field


class NameField(CharField, Field):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('max_length', 255)
        kwargs.setdefault('null', True)
        kwargs.setdefault('blank', True)
        kwargs.setdefault('db_index', True)
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["max_length"]
        del kwargs["db_index"]
        return name, path, args, kwargs

class NameStrictField(CharField, Field):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('max_length', 255)
        kwargs.setdefault('db_index', True)
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["max_length"]
        del kwargs["db_index"]
        return name, path, args, kwargs



import logging

from isc_common.fields.related import ForeignKeyCascade, ForeignKeyProtect
from isc_common.models.audit import AuditModel, AuditManager, AuditQuerySet
from tracker.models.messages_theme import Messages_theme
from twits.models.chat_user_user import Chat_user_user

logger = logging.getLogger(__name__)


class Chat_user_user_themeQuerySet(AuditQuerySet):
    def filter(self, *args, **kwargs):
        return super().filter(*args, **kwargs)


class Chat_user_user_themeManager(AuditManager):

    @staticmethod
    def getRecord(record):
        res = {
            'id': record.id,
        }
        return res

    def get_queryset(self):
        return Chat_user_user_themeQuerySet(self.model, using=self._db)


class Chat_user_user_theme(AuditModel):
    chat_user_user = ForeignKeyCascade(Chat_user_user)
    theme = ForeignKeyProtect(Messages_theme)

    objects = Chat_user_user_themeManager()

    def __str__(self):
        return f"{self.id}"

    class Meta:
        verbose_name = 'Кросс таблтца'
        unique_together = (('chat_user_user', 'theme'),)



from typing import Text


def getOrElse(default, value):
    if value:
        return value
    else:
        return default


def blinkString(text, blink=True, color="black", bold=False) -> Text:
    if blink:
        res = f'<div class="blink"><strong><font color="{color}"</font>{text}</strong></div>'
    else:
        res = f'<div><strong><font color="{color}"</font>{text}</strong></div>'

    if bold:
        return f'<b>{res}</b>'
    else:
        return res



import logging

from django.core.management import BaseCommand
from django.db import connection

from code_generator.core.writer import dbl_qutes_str

logger = logging.getLogger(__name__)


class Command(BaseCommand):

    def add_arguments(self, parser):
        parser.add_argument('--obj_name', type=str)
        parser.add_argument('--prefix', type=str)

    def handle(self, *args, **options):
        obj_name = options.get('obj_name')
        prefix = options.get('prefix', None)

        sql_text = '''SELECT a.attname
                        --        pg_catalog.format_type(a.atttypid, a.atttypmod),
                        --        a.attnotnull,
                        --        a.atthasdef,
                        --        a.attnum
                        FROM pg_catalog.pg_attribute a
                        WHERE a.attrelid = (
                            SELECT c.oid
                            FROM pg_catalog.pg_class c --Этот подзапрос возвращает --идентификатор таблицы...
                                     LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
                            WHERE pg_catalog.pg_table_is_visible(c.oid)
                              AND c.relname ~ %s
                        )
                          AND a.attnum > 0
                          AND NOT a.attisdropped
                        ORDER BY a.attnum
                        '''

        with connection.cursor() as cursor:
            cursor.execute(sql_text,[obj_name])
            rows = cursor.fetchall()
            columns = []
            for row in rows:
                column_name, = row
                if prefix:
                    columns.append(f'{prefix}.{dbl_qutes_str(column_name)}')
                else:
                    columns.append(dbl_qutes_str(column_name))

        res = ',\n'.join(columns)
        print(res)



from django.core import checks
from django.db import models


class Field(models.Field):
    def check(self, **kwargs):
        return [
            *super().check(**kwargs),
            *self._check_verbose_name(**kwargs),
        ]

    def _check_verbose_name(self, **kwargs):
        if self.verbose_name is None:
            return [
                checks.Error(
                    "Fields must define a 'verbose_name' attribute.",
                    obj=self,
                    id='fields.E120',
                )
            ]
        else:
            return []



import logging

from django.core.management import BaseCommand

from code_generator.core.writer import qutes_str, int_2_bin_str, reverse_str, fill_ch

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Создание уникальных ограничителей с учетом null столбцов"

    # constraints = [
    #     UniqueConstraint(fields=['STMP_2', 'props'], condition=Q(STMP_1=None) | Q(version=None), name='Item_unique_1'),
    #     UniqueConstraint(fields=['STMP_1', 'props'], condition=Q(STMP_2=None) | Q(version=None), name='Item_unique_2'),
    #     UniqueConstraint(fields=['STMP_2', 'version', 'props'], condition=Q(STMP_1=None), name='Item_unique_3'),
    #     UniqueConstraint(fields=['STMP_1', 'version', 'props'], condition=Q(STMP_2=None), name='Item_unique_4'),
    #     UniqueConstraint(fields=['STMP_1', 'STMP_2', 'props'], condition=Q(version=None), name='Item_unique_5'),
    #     UniqueConstraint(fields=['STMP_1', 'STMP_2', 'version', 'props'], name='Item_unique_6'),
    # ]

    def add_arguments(self, parser):
        parser.add_argument('--table', type=str)
        parser.add_argument('--columns', type=str)

    def handle(self, *args, **options):
        table = options.get('table')
        columns = options.get('columns')

        _columns = columns.split(',')
        not_null_column = []
        null_column = []
        constraints = []

        for column in _columns:
            _column = column.split('=')
            if len(_column) > 1 and _column[1] == 'null':
                null_column.append(_column[0].strip())
            else:
                not_null_column.append(qutes_str(_column[0].strip()))

        for i in range(0, 2 ** len(null_column)):
            condition = []
            b = reverse_str(int_2_bin_str(i, len(null_column)))
            idx = 0
            cl_null_column = []
            cl_not_null_column = sorted(not_null_column.copy())
            for _b in b:
                if _b == '1':
                    cl_not_null_column.append(qutes_str(null_column[idx]))
                else:
                    cl_null_column.append(null_column[idx])
                idx += 1

            for _null_column in cl_null_column:
                condition.append(f'Q({_null_column}=None)')

            not_null_column_str = ', '.join(sorted(cl_not_null_column))
            condition_str = ' & '.join(sorted(condition))
            if not condition_str:
                uconstraint = fill_ch(f'''UniqueConstraint(fields=[{not_null_column_str}], name='{table}_unique_constraint_{i}'),\n''', 4, ' ')
            else:
                uconstraint = fill_ch(f'''UniqueConstraint(fields=[{not_null_column_str}], condition={condition_str}, name='{table}_unique_constraint_{i}'),\n''', 4, ' ')
            constraints.append(uconstraint)
        res = f'''constraints = [\n{''.join(constraints)}]'''
        print(res)



import logging

from django.db import transaction

from isc_common import delAttr
from isc_common.auth.models.user import User
from isc_common.fields.related import ForeignKeyCascade
from isc_common.http.DSRequest import DSRequest
from isc_common.models.audit import AuditModel, AuditManager, AuditQuerySet
from reports.models.jasper_reports import Jasper_reports

logger = logging.getLogger(__name__)


class Jasper_reports_usersQuerySet(AuditQuerySet):
    pass


class Jasper_reports_usersManager(AuditManager):

    @staticmethod
    def getRecord(record):
        res = {
            'id': record.id,
            'user_id': record.user.id,
            'user__username': record.user.username,
            'user__first_name': record.user.first_name,
            'user__last_name': record.user.last_name,
            'user__email': record.user.email,
            'user__middle_name': record.user.middle_name,
        }
        return res

    def get_queryset(self):
        return Jasper_reports_usersQuerySet(self.model, using=self._db)

    def updateFromRequest(self, request, removed=None, function=None):
        request = DSRequest(request=request)
        data = request.get_data()
        _data = data.copy()

        delAttr(_data, 'id')
        delAttr(_data, 'user__username')
        delAttr(_data, 'user__first_name')
        delAttr(_data, 'user__last_name')
        delAttr(_data, 'user__email')
        delAttr(_data, 'user__middle_name')

        res = super().filter(id=data.get('id')).update(**_data)
        return data

    def createFromRequest(self, request):
        request = DSRequest(request=request)
        data = request.get_data()
        user_ids = data.get('user_ids', None)
        context_ids = data.get('context_ids', None)

        with transaction.atomic():
            if user_ids and context_ids:
                if not isinstance(user_ids, list):
                    user_ids = [user_ids]

                if not isinstance(context_ids, list):
                    context_ids = [context_ids]

                for user_id in user_ids:
                    for context_id in context_ids:
                        res, _ = super().get_or_create(user_id=user_id, report_id=context_id)

            return data


class Jasper_reports_users(AuditModel):
    report = ForeignKeyCascade(Jasper_reports)
    user = ForeignKeyCascade(User)

    objects = Jasper_reports_usersManager()

    def __str__(self):
        return f"ID:{self.id}"

    class Meta:
        verbose_name = 'Доступ пользователей к отчетам'



import json
import logging
import sys
import traceback

import websocket
from django.conf import settings


class WebSocket:
    procedures = []
    logger = logging.getLogger(__name__)

    @staticmethod
    def recv(**kwrags):
        host = kwrags.get('host')
        logger = kwrags.get('logger')

        if not host:
            raise Exception(f'host is not exists.')

        port = kwrags.get('port')
        if not port:
            raise Exception(f'port not exists.')

        if not isinstance(port, int):
            port = int(port)

        channel = kwrags.get('channel')
        if not channel:
            raise Exception(f'channel is not exists.')

        message = kwrags.get('message')

        if not message:
            raise Exception(f'message is not exists.')

        if not isinstance(message, dict):
            raise Exception(f'message is not dict instance')

        settings.LOCKS.acquire('WebSocket.recv')
        try:
            url = f'ws://{host}:{port}/ws/{channel}/'
            ws = websocket.create_connection(url)
            ws.send(json.dumps(message))
            settings.LOCKS.release('WebSocket.recv')
            return ws, json.loads(ws.recv())
        except Exception as ex:

            exc_info = sys.exc_info()
            message = str(ex)
            stackTrace = traceback.format_exception(*exc_info)

            logging.error(message)

            for x in stackTrace:
                logger.error(x)

            del exc_info
            settings.LOCKS.release('WebSocket.recv')
            return None

    @staticmethod
    def send_message(**kwrags):
        host = kwrags.get('host')
        logger = kwrags.get('logger')

        if not host:
            raise Exception(f'host is not exists.')

        port = kwrags.get('port')
        if not port:
            raise Exception(f'port not exists.')

        if not isinstance(port, int):
            port = int(port)

        channel = kwrags.get('channel')
        if not channel:
            raise Exception(f'channel is not exists.')

        message = kwrags.get('message')
        if not message:
            raise Exception(f'message is not exists.')

        if not isinstance(message, dict):
            raise Exception(f'message is not dict instance')

        try:
            settings.LOCKS.acquire('WebSocket.send_message')
            url = f'ws://{host}:{port}/ws/{channel}/'
            ws = websocket.create_connection(url)
            ws.send(json.dumps(message))
            ws.close()
            settings.LOCKS.release('WebSocket.send_message')
        except Exception as ex:

            exc_info = sys.exc_info()
            message = str(ex)
            stackTrace = traceback.format_exception(*exc_info)

            logging.error(message)

            for x in stackTrace:
                logger.error(x)

            del exc_info
            settings.LOCKS.release('WebSocket.send_message')

    def on_message(self, message):
        for proc in self.procedures:
            if hasattr(proc, '__call__'):
                proc(message)

    def on_error(self, error):
        self.logger.error(error)

    def on_close(self):
        self.logger.debug(f'### closed ###')

    def on_open(self):
        self.logger.debug(f'### opened ###')

    def remove_proc(self, proc):
        self.procedures.remove(proc)

    def append_proc(self, proc):
        self.procedures.append(proc)

    def send(self, data):
        if isinstance(data, dict):
            self.webSocket.send(json.dumps(data))
        else:
            raise Exception(f'data is not dict.')

    def send_logging(self, msg, type='logging'):
        data = dict(msg=msg, type=type)
        # self.webSocket.send(json.dumps(data))

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v() if callable(v) else v)

        websocket.enableTrace(True)
        self.webSocket = websocket.WebSocketApp(f'ws://{self.host}:{self.port}/ws/{self.channel}/',
                                                on_message=self.on_message,
                                                on_error=self.on_error,
                                                on_close=self.on_close,
                                                on_open=self.on_open
                                                )
        self.webSocket.run_forever()



# Generated by Django 3.0.3 on 2020-02-08 13:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0079_auto_20200206_0600'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='messages',
            unique_together=set(),
        ),
        migrations.AddConstraint(
            model_name='messages',
            constraint=models.UniqueConstraint(condition=models.Q(guid=None), fields=('checksum', 'state', 'theme', 'to_whom', 'user'), name='Messages_unique_constraint_0'),
        ),
        migrations.AddConstraint(
            model_name='messages',
            constraint=models.UniqueConstraint(fields=('checksum', 'guid', 'state', 'theme', 'to_whom', 'user'), name='Messages_unique_constraint_1'),
        ),
    ]



# Generated by Django 2.2.5 on 2019-09-16 17:59

import bitfield.models
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('twits', '0017_auto_20190913_1056'),
    ]

    operations = [
        migrations.AlterField(
            model_name='chat_users',
            name='props',
            field=bitfield.models.BitField((('compulsory_reading', 'Обязательное прочтение'), ('relevant', 'Актуальность'), ('readonly', 'Только чтение')), db_index=True, default=0),
        ),
    ]



import logging

from isc_common.fields.related import ForeignKeyCascade, ForeignKeyProtect
from isc_common.models.audit import AuditModel, AuditManager, AuditQuerySet
from tracker.models.messages import Messages
from tracker.models.messages_files import Messages_files

logger = logging.getLogger(__name__)


class Messages_files_refsQuerySet(AuditQuerySet):
    def filter(self, *args, **kwargs):
        return super().filter(*args, **kwargs)


class Messages_files_refsManager(AuditManager):

    @staticmethod
    def getRecord(record):
        res = {
            'id': record.id,
        }
        return res

    def get_queryset(self):
        return Messages_files_refsQuerySet(self.model, using=self._db)


class Messages_files_refs(AuditModel):
    message = ForeignKeyProtect(Messages)
    messages_file = ForeignKeyProtect(Messages_files)

    objects = Messages_files_refsManager()

    def __str__(self):
        return f"{self.id}"

    class Meta:
        verbose_name = 'Кросс таблица'
        unique_together = (('message', 'messages_file'),)



from isc_common.http.DSResponse import DSResponseUpdate, DSResponseAdd, DSResponse, JsonResponseWithException
from isc_common.http.RPCResponse import RPCResponseConstant
from isc_common.http.response import JsonResponse
from tracker.models.messages_theme import Messages_theme
from tracker.models.messages_theme_admin_view import Messages_theme_admin_view, Messages_theme_admin_view_Manager
from tracker.models.messages_theme_view import Messages_theme_view, Messages_theme_view_Manager


@JsonResponseWithException(printing=False, printing_res=False)
def Messages_theme_Fetch(request):
    return JsonResponse(
        DSResponse(
            request=request,
            data=Messages_theme_view.objects.
                filter().
                values('code',
                       'creator_id',
                       'deleted_at',
                       'deliting',
                       'description',
                       'editing',
                       'id',
                       'isFolder',
                       'lastmodified',
                       'name',
                       'parent_id', ).
                distinct('id').
                get_range_rows1(
                request=request,
                function=Messages_theme_view_Manager.getRecord
            ),
            status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException(printing=False, printing_res=False)
def Messages_theme_FetchAdmin(request):
    return JsonResponse(
        DSResponse(
            request=request,
            data=Messages_theme_admin_view.objects.
                filter().
                get_range_rows1(
                request=request,
                function=Messages_theme_admin_view_Manager.getRecord
            ),
            status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Messages_theme_Add(request):
    return JsonResponse(DSResponseAdd(data=Messages_theme.objects.createFromRequest(request=request), status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Messages_theme_Update(request):
    return JsonResponse(DSResponseUpdate(data=Messages_theme.objects.updateFromRequest(request), status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Messages_theme_Remove(request):
    return JsonResponse(DSResponse(request=request, data=Messages_theme.objects.deleteFromRequest(request=request), status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Messages_theme_Lookup(request):
    return JsonResponse(DSResponse(request=request, data=Messages_theme.objects.lookupFromRequest(request=request), status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Messages_theme_Info(request):
    return JsonResponse(DSResponse(request=request, data=Messages_theme_view.objects.get_queryset().get_info(request=request), status=RPCResponseConstant.statusSuccess).response)



# Generated by Django 2.2.5 on 2019-09-05 05:19

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0058_auto_20190905_0517'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='messages',
            options={'ordering': ('date_create',), 'verbose_name': 'Сообщения'},
        ),
    ]



# Generated by Django 2.2.5 on 2019-09-09 13:57

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('twits', '0014_auto_20190909_1042'),
    ]

    operations = [
        migrations.AlterModelTable(
            name='chat_usersview',
            table='twits_chat_users_view',
        ),
    ]



import logging

from django.db.models import BooleanField

from isc_common import delAttr
from isc_common.auth.models.user import User
from isc_common.fields.related import ForeignKeyCascade
from isc_common.http.DSRequest import DSRequest
from isc_common.managers.common_tree_grid_manager import CommonTreeGridManager
from isc_common.models.base_ref import BaseRefHierarcy

logger = logging.getLogger(__name__)


class Messages_theme_admin_view_Manager(CommonTreeGridManager):
    @staticmethod
    def getRecord(record):
        res = {
            "id": record.id,
            "code": record.code,
            "name": record.name,
            # "full_name": record.full_name,
            "description": record.description,
            "parent_id": record.parent_id,
            "lastmodified": record.lastmodified,
            "isFolder": record.isFolder,
            "editing": record.editing,
            "deliting": record.deliting,
        }
        return res

class Messages_theme_admin_view(BaseRefHierarcy):
    creator = ForeignKeyCascade(User, null=True, blank=True, related_name='creator_theme_admin')
    isFolder = BooleanField(default=False)

    def __str__(self):
        return f"{self.code}: {self.name}"

    objects = Messages_theme_admin_view_Manager()

    class Meta:
        verbose_name = 'Темы задач'
        managed = False
        db_table = 'tracker_messages_theme_admin_view'



# Generated by Django 2.2.5 on 2019-09-08 06:10

from django.db import migrations
import django.db.models.deletion
import isc_common.fields.related


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0061_auto_20190907_0443'),
    ]

    operations = [
        migrations.AlterField(
            model_name='messages_files_refs',
            name='message',
            field=isc_common.fields.related.ForeignKeyProtect(on_delete=django.db.models.deletion.PROTECT, to='tracker.Messages'),
        ),
        migrations.AlterField(
            model_name='messages_files_refs',
            name='messages_file',
            field=isc_common.fields.related.ForeignKeyProtect(on_delete=django.db.models.deletion.PROTECT, to='tracker.Messages_files'),
        ),
    ]



# Generated by Django 2.2.7 on 2019-11-16 12:37

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('history', '0007_bannedip'),
    ]

    operations = [
        migrations.DeleteModel(
            name='BannedIP',
        ),
        migrations.DeleteModel(
            name='UntrackedUserAgent',
        ),
    ]



from isc_common.auth.http.ChangePassRequest import ChangePassRequest
from isc_common.auth.http.Link2Group import Link2Group
from isc_common.auth.http.LoginRequets import LoginRequest
from isc_common.auth.http.LogoutRequets import LogoutRequest
from isc_common.auth.http.UnLinkFromGroup import UnlinkFromGroup
from isc_common.auth.http.copyPermission import CopyPermission
from isc_common.http.DSResponse import JsonResponseWithException
from isc_common.http.response import JsonResponse


@JsonResponseWithException(printing=False)
def login(request):
    return JsonResponse(LoginRequest(request).response)


@JsonResponseWithException(printing=False)
def logout(request):
    return JsonResponse(LogoutRequest(request).response)


@JsonResponseWithException(printing=False)
def changepassword(request):
    return JsonResponse(ChangePassRequest(request).response)


@JsonResponseWithException(printing=False)
def link2group(request):
    return JsonResponse(Link2Group(request).response)


@JsonResponseWithException(printing=False)
def unlinkFromgroup(request):
    return JsonResponse(UnlinkFromGroup(request).response)


@JsonResponseWithException(printing=False)
def copyPermission(request):
    return JsonResponse(CopyPermission(request).response)



import datetime
import logging

from django.db import transaction
from django.forms import model_to_dict

from clndr.models.day import Day
from clndr.models.day_types import Day_types
from isc_common import delAttr, setAttr
from isc_common.datetime import StrToDate
from isc_common.fields.description_field import DescriptionField
from isc_common.fields.related import ForeignKeyProtect
from isc_common.fields.time_field import TimeField
from isc_common.http.DSRequest import DSRequest
from isc_common.managers.common_managet_with_lookup_fields import CommonManagetWithLookUpFieldsManager, CommonManagetWithLookUpFieldsQuerySet
from isc_common.models.audit import AuditModel

logger = logging.getLogger(__name__)


class Shift_dayQuerySet(CommonManagetWithLookUpFieldsQuerySet):
    def filter(self, *args, **kwargs):
        return super().filter(*args, **kwargs)


class Shift_dayManager(CommonManagetWithLookUpFieldsManager):

    @staticmethod
    def getRecord(record):
        res = {
            'id': record.id,

            'day_id': record.day.id,
            'day__code': record.day.code,
            'day__name': record.day.name,
            'day__description': record.day.description,

            'daytype_id': record.daytype.id,
            'daytype__code': record.daytype.code,
            'daytype__name': record.daytype.name,
            'daytype__color': record.daytype.color,
            'daytype__length': record.daytype.length,

            'color': record.daytype.color,
            'begintime': record.begintime,
            'endtime': record.endtime,
            'description': record.description,

            'info': record.info,
        }
        return res

    def get_queryset(self):
        return Shift_dayQuerySet(self.model, using=self._db)

    def createFromRequest(self, request):

        request = DSRequest(request=request)
        data = request.get_data()

        return self.create_update(data)

    def updateFromRequest(self, request):

        request = DSRequest(request=request)
        data = request.get_data()

        return self.create_update(data)

    def copyFromRequest(self, request):

        request = DSRequest(request=request)
        data = request.get_data()
        id = data.get('id')

        shiftday = Shift_day.objects.get(id = id)
        shiftday = shiftday.copy()

        _shiftday = model_to_dict(shiftday)
        setAttr(_shiftday, 'info', shiftday.info)
        setAttr(_shiftday, 'color', shiftday.color)

        return _shiftday

    def create_update(self, data):
        from clndr.models.shift_day_relax import Shift_day_relax
        shiftdays_relax = data.get('shiftdays_relax')
        delAttr(data, 'shiftdays_relax')
        delAttr(data, 'color')
        delAttr(data, 'info')
        id = data.get('id')
        delAttr(data, 'id')

        with transaction.atomic():
            delAttr(data, 'day')
            delAttr(data, 'daytype')
            delAttr(data, 'daytype')
            delAttr(data, 'day__name')
            delAttr(data, 'daytype__name')

            shiftday, _ = super().update_or_create(id=id, defaults=data)
            Shift_day_relax.objects.filter(shiftday_id=shiftday.id).delete()

            if isinstance(shiftdays_relax, list):
                for shiftday_relax in shiftdays_relax:
                    delAttr(shiftday_relax, 'id')
                    shiftday_relax = self._remove_prop_(shiftday_relax)
                    setAttr(shiftday_relax, 'shiftday', shiftday)
                    Shift_day_relax.objects.create(**shiftday_relax)

            setAttr(data, 'id', shiftday.id)
            setAttr(data, 'color', shiftday.color)
            setAttr(data, 'info', shiftday.info)
            setAttr(data, 'day__name', shiftday.day.name)
            setAttr(data, 'begintime', StrToDate(shiftday.begintime,'%H:%M:%S.%f').time())
            setAttr(data, 'endtime', StrToDate(shiftday.endtime,'%H:%M:%S.%f').time())
            setAttr(data, 'daytype__name', shiftday.daytype.name)
            return data


class Shift_day(AuditModel):
    description = DescriptionField()

    day = ForeignKeyProtect(Day)
    daytype = ForeignKeyProtect(Day_types)

    begintime = TimeField(verbose_name='Начало', db_index=True)
    endtime = TimeField(verbose_name='Конец', db_index=True)

    @property
    def name(self):
        return self.day.name

    def copy(self):
        from clndr.models.shift_day_relax import Shift_day_relax

        data = model_to_dict(Shift_day.objects.get(id=self.id))
        delAttr(data, 'id')

        day = data.get('day')
        delAttr(data, 'day')
        setAttr(data, 'day_id', day)

        daytype = data.get('daytype')
        delAttr(data, 'daytype')
        setAttr(data, 'daytype_id', daytype)

        with transaction.atomic():
            shiftday = Shift_day.objects.create(**data)

            for shift_day_relax in Shift_day_relax.objects.filter(shiftday_id=self.id):
                shift_day_relax = model_to_dict(shift_day_relax)
                delAttr(shift_day_relax, 'id')
                setAttr(shift_day_relax, 'shiftday', shiftday)
                Shift_day_relax.objects.create(**shift_day_relax)

            return shiftday

    @property
    def day_html(self):
        return f'<font color="{self.daytype.color}"><b>{self.day.name} ({self.daytype.name})</b></font>'

    @property
    def period_html(self):
        if self.begintime == self.endtime:
            return ''
        return f'<p><b>{self.begintime} - {self.endtime}</b></p>'

    @property
    def color(self):
        return self.daytype.color

    @property
    def info(self):
        from clndr.models.shift_day_relax import Shift_day_relax
        res = self.day_html
        res += self.period_html

        for sd_rx in Shift_day_relax.objects.filter(shiftday=self).order_by(*['relaxorder']):
            res += sd_rx.html
        return f'<div style="border:1px solid grey; merging: 2px; padding: 5px">{res}</div>'

    objects = Shift_dayManager()

    def __str__(self):
        return f"ID={self.id}, " \
               f"\n  description={self.description}, " \
               f"\n  day=[{self.day}], " \
               f"\n  daytype=[{self.daytype}], " \
               f"\n  begintime={self.begintime}, " \
               f"\n  endtime={self.endtime}"

    class Meta:
        verbose_name = 'День смены'



import calendar
import logging
from datetime import datetime

from django.core.management import BaseCommand

from clndr.calendar_event import CalendarEvent, CalendarEventLinkedList
from clndr.models.holidays import Holidays

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = 'Создание модели календаря.'

    shifts = CalendarEventLinkedList()
    relaxes = CalendarEventLinkedList()

    def handle(self, *args, **options):

        self.red_label_days = CalendarEventLinkedList(query=Holidays.objects.filter(daytype__isholiday=True).order_by(*['begindate']))
        self.red_label_days.print('red_label_days')

        self.shorted_days = CalendarEventLinkedList(query=Holidays.objects.exclude(daytype__length__isnull=True))
        self.shorted_days.print('shorted_days')

        year = 2019
        for month in range(1, 12):
            _, month_range = calendar.monthrange(year, month)
            for day in range(1, month_range):
                weekday = calendar.weekday(year, month, day)

                if weekday < 5:
                    item = datetime(year=year, month=month, day=day, hour=7)
                    red_label_day = self.red_label_days.find(item, lambda item, x: x.begindate <= item and x.enddate > item)

                    if not red_label_day:
                        if month_range < day + 1:
                            if month + 1 > 12:
                                self.shifts.add(CalendarEvent(
                                    startDate=datetime(year=year, month=month, day=day, hour=16),
                                    endDate=datetime(year=year + 1, month=1, day=1, hour=1),
                                    iswork=True
                                ))
                            else:
                                self.shifts.add(CalendarEvent(
                                    startDate=datetime(year=year, month=month, day=day, hour=16),
                                    endDate=datetime(year=year, month=month + 1, day=1, hour=1),
                                    iswork=True
                                ))
                        else:
                            self.shifts.add(CalendarEvent(
                                startDate=datetime(year=year, month=month, day=day, hour=16),
                                endDate=datetime(year=year, month=month, day=day + 1, hour=1),
                                iswork=True
                            ))

                        self.relaxes.add(CalendarEvent(
                            startDate=datetime(year=year, month=month, day=day, hour=9),
                            endDate=datetime(year=year, month=month, day=day, hour=9, minute=15),

                        ))

                        self.relaxes.add(CalendarEvent(
                            startDate=datetime(year=year, month=month, day=day, hour=11),
                            endDate=datetime(year=year, month=month, day=day, hour=12),

                        ))

                        self.relaxes.add(CalendarEvent(
                            startDate=datetime(year=year, month=month, day=day, hour=14),
                            endDate=datetime(year=year, month=month, day=day, hour=14, minute=15),

                        ))

                        self.relaxes.add(CalendarEvent(
                            startDate=datetime(year=year, month=month, day=day, hour=18),
                            endDate=datetime(year=year, month=month, day=day, hour=18, minute=15),

                        ))

                        self.relaxes.add(CalendarEvent(
                            startDate=datetime(year=year, month=month, day=day, hour=20),
                            endDate=datetime(year=year, month=month, day=day, hour=21),

                        ))

                        self.relaxes.add(CalendarEvent(
                            startDate=datetime(year=year, month=month, day=day, hour=23),
                            endDate=datetime(year=year, month=month, day=day, hour=23, minute=15),

                        ))
                    else:
                        if month_range < day + 1:
                            if month + 1 > 12:
                                self.shifts.add(CalendarEvent(
                                    startDate=red_label_day.begindate,
                                    endDate=datetime(year=red_label_day.begindate.year + 1, month=1, day=1),
                                    name=red_label_day.name,
                                    isredlabelday=True
                                ))
                            else:
                                self.shifts.add(CalendarEvent(
                                    startDate=red_label_day.begindate,
                                    endDate=datetime(year=red_label_day.begindate.year, month=red_label_day.begindate.month + 1, day=1),
                                    name=red_label_day.name,
                                    isredlabelday=True
                                ))
                        else:
                            self.shifts.add(CalendarEvent(
                                startDate=red_label_day.begindate,
                                endDate=datetime(year=red_label_day.begindate.year, month=red_label_day.begindate.month, day=red_label_day.begindate.day + 1),
                                name=red_label_day.name,
                                isredlabelday=True
                            ))
                else:
                    if month_range < day + 1:
                        if month + 1 > 12:
                            self.shifts.add(CalendarEvent(
                                startDate=datetime(year=year, month=month, day=day),
                                endDate=datetime(year=year + 1, month=1, day=1),
                                iswork=True
                            ))
                        else:
                            self.shifts.add(CalendarEvent(
                                startDate=datetime(year=year, month=month, day=day),
                                endDate=datetime(year=year, month=month + 1, day=1),
                                iswork=True
                            ))
                    else:
                        self.shifts.add(CalendarEvent(
                            startDate=datetime(year=year, month=month, day=day),
                            endDate=datetime(year=year, month=month, day=day + 1),
                            iswork=True
                        ))


        self.shifts.fusion(self.relaxes)
        self.shifts.print('shifts')



from datetime import datetime


def str2Date(date_str):
    if date_str is None:
        return None
    else:
        return datetime.strptime(date_str, '%Y-%m-%dT%I:%M:%S.%fZ').date()


def str2Date4Django(date_str):
    if date_str is None:
        return None
    else:
        date = str2Date(date_str)
        if (date.month < 10):
            month = f"0{date.month}"
        else:
            month = date.month

        if (date.day < 10):
            day = f"0{date.day}"
        else:
            day = date.day

        return f"{date.year}-{month}-{day}"



import logging
import sys
import uuid
from typing import Text

from django.conf import settings
from django.db import connection
from isc_common.common import blinkString

logger = logging.getLogger(__name__)


def create_tmp_mat_view(sql_str, params=[], indexes=[], mat_view_name=None) -> Text:
    if mat_view_name != None:
        settings.LOCKS.acquire(f'create_tmp_mat_view_{mat_view_name}')
    else:
        mat_view_name = f'tmp_{str(uuid.uuid4()).replace("-", "_")}'

    if isinstance(indexes, list):
        indexes = [f'''CREATE INDEX {mat_view_name}_{index}_idx ON {mat_view_name} USING btree ("{index}")''' for index in indexes]
        suffix = ';'.join(indexes)
    sql_txt = f'''CREATE MATERIALIZED VIEW {mat_view_name} AS {sql_str} WITH DATA; {suffix}'''
    try:
        with connection.cursor() as cursor:
            logger.debug(f'Creating: {mat_view_name}')
            cursor.execute(sql_txt, params)
            logger.debug(f'Created: {mat_view_name}')
        if mat_view_name != None:
            settings.LOCKS.release(f'create_tmp_mat_view_{mat_view_name}')
        return mat_view_name
    except Exception as ex:
        exc_info = sys.exc_info()
        logger.error(msg=ex, exc_info=exc_info)
        if mat_view_name != None:
            settings.LOCKS.release(f'create_tmp_mat_view_{mat_view_name}')
        return None


def create_view(sql_str, params=[], view_name=None) -> None:
    if view_name == None:
        raise Exception(f'View must have name/')
    settings.LOCKS.acquire(f'createt_view_{view_name}')

    sql_txt = f'''DROP VIEW IF EXISTS {view_name} CASCADE;'''
    sql_txt += f'''CREATE VIEW {view_name} AS {sql_str};'''
    try:
        with connection.cursor() as cursor:
            logger.debug(f'Creating: {view_name}')
            cursor.execute(sql_txt, params)
            logger.debug(f'Created: {view_name}')
        settings.LOCKS.release(f'createt_view_{view_name}')
    except Exception as ex:
        exc_info = sys.exc_info()
        logger.error(msg=ex, exc_info=exc_info)
        settings.LOCKS.release(f'createt_view_{view_name}')


def create_tmp_table(fields_str=None, sql_str=None, on_commit='delete rows', params=[], indexes=[], unique_indexes=[], table_name=None, drop=True) -> None:
    if table_name == None:
        raise Exception(f'Table must have name/')
    settings.LOCKS.acquire(f'create_tmp_table{table_name}')

    indexes_str = ''
    if isinstance(indexes, list):
        indexes = [f'''CREATE INDEX {table_name}_{index}_idx ON {table_name} USING btree ("{index}")''' for index in indexes]
        indexes_str = ';'.join(indexes)

    unique_indexes_str = ''
    if isinstance(unique_indexes, list):
        indexes = [f'''CREATE INDEX {table_name}_{index[0]}_idx ON {table_name} USING btree ("{index[0]}") WHERE ({index[1] if index[1] else "1=1"});''' for index in unique_indexes]
        unique_indexes_str = ';'.join(indexes)

    sql_txt = ''
    if drop:
        sql_txt = f'''DROP table IF EXISTS {table_name} CASCADE;'''

    on_commit_str = f'ON COMMIT {on_commit}'
    if on_commit == None:
        on_commit_str = ''

    if sql_str == None:
        sql_txt += f'''CREATE TEMPORARY TABLE {table_name} ({fields_str}) {on_commit_str};'''
    else:
        sql_txt += f'''CREATE TEMPORARY TABLE {table_name} as ({sql_str}); {on_commit_str};'''
    sql_txt += indexes_str
    sql_txt += unique_indexes_str

    try:
        with connection.cursor() as cursor:
            logger.debug(f'Creating: {table_name}')
            cursor.execute(sql_txt, params)
            logger.debug(f'Created: {table_name}')
        settings.LOCKS.release(f'create_tmp_table{table_name}')
    except Exception as ex:
        exc_info = sys.exc_info()
        logger.error(msg=ex, exc_info=exc_info)
        settings.LOCKS.release(f'create_tmp_table{table_name}')


def drop_mat_view(mat_view_name) -> bool:
    sql_txt = f'DROP MATERIALIZED VIEW IF EXISTS {mat_view_name} CASCADE;'
    try:
        settings.LOCKS.acquire(f'drop_mat_view_{mat_view_name}')
        with connection.cursor() as cursor:
            logger.debug(f'Droping: {mat_view_name}')
            cursor.execute(sql_txt)
        settings.LOCKS.release(f'drop_mat_view_{mat_view_name}')
        logger.debug(f'Droped: {mat_view_name}')
        return True
    except Exception as ex:
        exc_info = sys.exc_info()
        logger.error(msg=ex, exc_info=exc_info)
        settings.LOCKS.release(f'drop_mat_view_{mat_view_name}')
        return False


def refresh_mat_view(mat_view_name, progressStack=None):
    sql_txt = f'REFRESH MATERIALIZED VIEW {mat_view_name};'
    settings.LOCKS.acquire(f'refresh_mat_view_{mat_view_name}')
    try:
        with connection.cursor() as cursor:
            if progressStack != None:
                progressStack.setContentsLabel(blinkString(f'Обновление материализованного преставления {mat_view_name}', bold=True))
            cursor.execute(sql_txt)
        settings.LOCKS.release(f'refresh_mat_view_{mat_view_name}')
        if progressStack != None:
            progressStack.close()
        return True
    except Exception as ex:
        exc_info = sys.exc_info()
        logger.error(msg=ex, exc_info=exc_info)
        settings.LOCKS.release(f'refresh_mat_view_{mat_view_name}')
        if progressStack != None:
            progressStack.close()
        return False



from django.urls import path

from isc_common.auth.views import user

urlpatterns = [

    path('User/Fetch/', user.User_Fetch),
    path('User/FetchExclBots/', user.User_FetchExclBots),
    path('User/Add', user.User_Add),
    path('User/Update', user.User_Update),
    path('User/Remove', user.User_Remove),
    path('User/Lookup/', user.User_Lookup),
    path('User/Info/', user.User_Info),

]



class RPCRequest:
    def __init__(self, _dict):
        if _dict:
            for key, value in _dict.items():
                if isinstance(value, (list, tuple)):
                   setattr(self, key, [RPCRequest(x) if isinstance(x, dict) else x for x in value])
                else:
                   setattr(self, key, RPCRequest(value) if isinstance(value, dict) else value)



# Generated by Django 2.2.4 on 2019-08-21 17:21

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('clndr', '0017_holidays_day_type'),
    ]

    operations = [
        migrations.RenameField(
            model_name='holidays',
            old_name='day_type',
            new_name='daytype',
        ),
    ]



import logging
import xml
from xml.dom import Node
from xml.dom.minicompat import NodeList

logger = logging.getLogger(__name__)

def _get_elements_by_part_tagName_helper(parent, name, rc):
    for node in parent.childNodes:
        if node.nodeType == Node.ELEMENT_NODE:
            logging.debug(f' node.tagName: {node.tagName}')
            if (name == "*" or node.tagName == name or node.tagName.find(name) != -1):
                rc.append(node)
        _get_elements_by_part_tagName_helper(node, name, rc)
    return rc


class Document(xml.dom.minidom.Document):
    def getElementsByPartTagName(self, name):
        return _get_elements_by_part_tagName_helper(self, name, NodeList())



"""
Windows Management Instrumentation (WMI) is Microsoft's answer to
the DMTF's Common Information Model. It allows you to query just
about any conceivable piece of information from any computer which
is running the necessary agent and over which have you the
necessary authority.

Since the COM implementation doesn't give much away to Python
programmers, I've wrapped it in some lightweight classes with
some getattr / setattr magic to ease the way. In particular:

* The :class:`_wmi_namespace` object itself will determine its classes
  and allow you to return all instances of any of them by
  using its name as an attribute::

    disks = wmi.WMI ().Win32_LogicalDisk ()

* In addition, you can specify what would become the WHERE clause
  as keyword parameters::

    fixed_disks = wmi.WMI ().Win32_LogicalDisk (DriveType=3)

* The objects returned by a WMI lookup are wrapped in a Python
  class which determines their methods and classes and allows
  you to access them as though they were Python classes. The
  methods only allow named parameters::

    for p in wmi.WMI ().Win32_Process (Name="notepad.exe"):
      p.Terminate (Result=1)

* Doing a print on one of the WMI objects will result in its
  `GetObjectText\_` method being called, which usually produces
  a meaningful printout of current values.
  The repr of the object will include its full WMI path,
  which lets you get directly to it if you need to.

* You can get the associators and references of an object as
  a list of python objects by calling the associators () and
  references () methods on a WMI Python object::

    for p in wmi.WMI ().Win32_Process (Name="notepad.exe"):
      for r in p.references ():
        print r

  ..  note::
      Don't do this on a Win32_ComputerSystem object; it will
      take all day and kill your machine!


* WMI classes (as opposed to instances) are first-class
  objects, so you can get hold of a class, and call
  its methods or set up a watch against it::

    process = wmi.WMI ().Win32_Process
    process.Create (CommandLine="notepad.exe")

* To make it easier to use in embedded systems and py2exe-style
  executable wrappers, the module will not force early Dispatch.
  To do this, it uses a handy hack by Thomas Heller for easy access
  to constants.

Typical usage will be::

  import wmi

  vodev1 = wmi.WMI ("vodev1")
  for disk in vodev1.Win32_LogicalDisk ():
    if disk.DriveType == 3:
      space = 100 * long (disk.FreeSpace) / long (disk.Size)
      print "%s has %d%% free" % (disk.Name, space)

Many thanks, obviously to Mark Hammond for creating the win32all
extensions, but also to Alex Martelli and Roger Upole, whose
c.l.py postings pointed me in the right direction.
Thanks especially in release 1.2 to Paul Tiemann for his code
contributions and robust testing.
"""
__VERSION__ = __version__ = "1.4.9"

_DEBUG = False

import datetime
import re
import struct
import sys
import warnings

import pywintypes
from win32com.client import GetObject, Dispatch


def signed_to_unsigned(signed):
    """Convert a (possibly signed) long to unsigned hex. Useful
    when converting a COM error code to the more conventional
    8-digit hex::

      print "%08X" % signed_to_unsigned (-2147023174)
    """
    unsigned, = struct.unpack("L", struct.pack("l", signed))
    return unsigned


class SelfDeprecatingDict(object):
    """Provides for graceful degradation of objects which
    are currently dictionaries (and therefore accessed via
    `.keys`, `.items`, etc.) into lists. Wraps an existing
    `dict` and allows it to be addressed as a `dict` or as a
    `list` during an interregnum, issuing a `DeprecationWarning`
    if accessed as a `dict`.
    """

    dict_only = set(dir(dict)).difference(dir(list))

    def __init__(self, dictlike):
        self.dict = dict(dictlike)
        self.list = list(self.dict)

    def __getattr__(self, attribute):
        if attribute in self.dict_only:
            warnings.warn("In future this will be a list and not a dictionary", DeprecationWarning)
            return getattr(self.dict, attribute)
        else:
            return getattr(self.list, attribute)

    def __iter__(self):
        return iter(self.list)

    def __str__(self):
        return str(self.list)

    def __repr__(self):
        return repr(self.list)

    def __getitem__(self, item):
        try:
            return self.list[item]
        except TypeError:
            return self.dict[item]


class ProvideConstants(object):
    """When called on a ``win32com.client.Dispatch`` object,
    provides lazy access to constants defined in the typelib.
    They can then be accessed as attributes of the :attr:`_constants`
    property. (From Thomas Heller on c.l.py).
    """

    def __init__(self, comobj):
        comobj.__dict__["_constants"] = self
        self.__typecomp = \
            comobj._oleobj_.GetTypeInfo().GetContainingTypeLib()[0].GetTypeComp()

    def __getattr__(self, name):
        if name.startswith("__") and name.endswith("__"):
            raise AttributeError(name)
        result = self.__typecomp.Bind(name)
        if not result[0]:
            raise AttributeError(name)
        return result[1].value


obj = GetObject("winmgmts:")
ProvideConstants(obj)

wbemErrInvalidQuery = obj._constants.wbemErrInvalidQuery
wbemErrTimedout = obj._constants.wbemErrTimedout
wbemFlagReturnImmediately = obj._constants.wbemFlagReturnImmediately
wbemFlagForwardOnly = obj._constants.wbemFlagForwardOnly


#
# Exceptions
#
class x_wmi(Exception):
    """Ancestor of all wmi-related exceptions. Keeps track of
    an info message and the underlying COM error if any, exposed
    as the :attr:`com_error` attribute.
    """

    def __init__(self, info="", com_error=None):
        self.info = info
        self.com_error = com_error

    def __str__(self):
        return "<x_wmi: %s %s>" % (
            self.info or "Unexpected COM Error",
            self.com_error or "(no underlying exception)"
        )


class x_wmi_invalid_query(x_wmi):
    "Raised when a WMI returns `wbemErrInvalidQuery`"
    pass


class x_wmi_timed_out(x_wmi):
    "Raised when a watcher times out"
    pass


class x_wmi_no_namespace(x_wmi):
    """Raised when an attempt is made to query or watch
    from a class without a namespace.
    """
    pass


class x_access_denied(x_wmi):
    "Raised when WMI raises 80070005"
    pass

class x_wmi_authentication(x_wmi):
    "Raised when an invalid combination of authentication properties is attempted when connecting"
    pass


class x_wmi_uninitialised_thread(x_wmi):
    """Raised when WMI returns 800401E4 on connection, usually
    indicating that no COM threading model has been initialised
    """
    pass


WMI_EXCEPTIONS = {
    signed_to_unsigned(wbemErrInvalidQuery): x_wmi_invalid_query,
    signed_to_unsigned(wbemErrTimedout): x_wmi_timed_out,
    2147944126: None,
    0x80070005: x_access_denied,
    0x80041003: x_access_denied,
    0x800401E4: x_wmi_uninitialised_thread,
}


def handle_com_error(err=None):
    """Convenience wrapper for displaying all manner of COM errors.
    Raises a :exc:`x_wmi` exception with more useful information attached

    :param err: The structure attached to a `pywintypes.com_error`
    """
    if err is None:
        _, err, _ = sys.exc_info()
    hresult_code, hresult_name, additional_info, parameter_in_error = err.args
    hresult_code = signed_to_unsigned(hresult_code)
    exception_string = ["%s - %s" % (hex(hresult_code), hresult_name)]
    scode = None
    if additional_info:
        wcode, source_of_error, error_description, whlp_file, whlp_context, scode = additional_info
        scode = signed_to_unsigned(scode)
        exception_string.append("  Error in: %s" % source_of_error)
        exception_string.append("  %s - %s" % (hex(scode), (error_description or "").strip()))
    for error_code, klass in WMI_EXCEPTIONS.items():
        if error_code in (hresult_code, scode):
            return  hresult_code, klass,  exception_string
    else:
        klass = x_wmi
        return hresult_code, klass, exception_string
    # raise klass(com_error=err)


BASE = datetime.datetime(1601, 1, 1)


def from_1601(ns100):
    return BASE + datetime.timedelta(microseconds=int(ns100) / 10)


def from_time(year=None, month=None, day=None, hours=None, minutes=None, seconds=None, microseconds=None, timezone=None):
    """Convenience wrapper to take a series of date/time elements and return a WMI time
    of the form `yyyymmddHHMMSS.mmmmmm+UUU`. All elements may be int, string or
    omitted altogether. If omitted, they will be replaced in the output string
    by a series of stars of the appropriate length.

    :param year: The year element of the date/time
    :param month: The month element of the date/time
    :param day: The day element of the date/time
    :param hours: The hours element of the date/time
    :param minutes: The minutes element of the date/time
    :param seconds: The seconds element of the date/time
    :param microseconds: The microseconds element of the date/time
    :param timezone: The timeezone element of the date/time

    :returns: A WMI datetime string of the form: `yyyymmddHHMMSS.mmmmmm+UUU`
    """

    def str_or_stars(i, length):
        if i is None:
            return "*" * length
        else:
            return str(i).rjust(length, "0")

    wmi_time = ""
    wmi_time += str_or_stars(year, 4)
    wmi_time += str_or_stars(month, 2)
    wmi_time += str_or_stars(day, 2)
    wmi_time += str_or_stars(hours, 2)
    wmi_time += str_or_stars(minutes, 2)
    wmi_time += str_or_stars(seconds, 2)
    wmi_time += "."
    wmi_time += str_or_stars(microseconds, 6)
    if timezone >= 0:
        wmi_time += "+"
    else:
        wmi_time += "-"
        timezone = abs(timezone)
    wmi_time += str_or_stars(timezone, 3)

    return wmi_time


def to_time(wmi_time):
    """Convenience wrapper to take a WMI datetime string of the form
    yyyymmddHHMMSS.mmmmmm+UUU and return a 9-tuple containing the
    individual elements, or None where string contains placeholder
    stars.

    :param wmi_time: The WMI datetime string in `yyyymmddHHMMSS.mmmmmm+UUU` format

    :returns: A 9-tuple of (year, month, day, hours, minutes, seconds, microseconds, timezone)
    """

    def int_or_none(s, start, end):
        try:
            return int(s[start:end])
        except ValueError:
            return None

    year = int_or_none(wmi_time, 0, 4)
    month = int_or_none(wmi_time, 4, 6)
    day = int_or_none(wmi_time, 6, 8)
    hours = int_or_none(wmi_time, 8, 10)
    minutes = int_or_none(wmi_time, 10, 12)
    seconds = int_or_none(wmi_time, 12, 14)
    microseconds = int_or_none(wmi_time, 15, 21)
    timezone = wmi_time[22:]
    if timezone == "***":
        timezone = None

    return year, month, day, hours, minutes, seconds, microseconds, timezone


def _set(obj, attribute, value):
    """Helper function to add an attribute directly into the instance
    dictionary, bypassing possible `__getattr__` calls

    :param obj: Any python object
    :param attribute: String containing attribute name
    :param value: Any python object
    """
    obj.__dict__[attribute] = value


class _wmi_method:
    """A currying sort of wrapper around a WMI method name. It
    abstract's the method's parameters and can be called like
    a normal Python object passing in the parameter values.

    Output parameters are returned from the call as a tuple.
    In addition, the docstring is set up as the method's
    signature, including an indication as to whether any
    given parameter is expecting an array, and what
    special privileges are required to call the method.
    """

    def __init__(self, ole_object, method_name):
        """
        :param ole_object: The WMI class/instance whose method is to be called
        :param method_name: The name of the method to be called
        """
        try:
            self.ole_object = Dispatch(ole_object)
            self.method = ole_object.Methods_(method_name)
            self.qualifiers = {}
            for q in self.method.Qualifiers_:
                self.qualifiers[q.Name] = q.Value
            self.provenance = "\n".join(self.qualifiers.get("MappingStrings", []))

            self.in_parameters = self.method.InParameters
            self.out_parameters = self.method.OutParameters
            if self.in_parameters is None:
                self.in_parameter_names = []
            else:
                self.in_parameter_names = [(i.Name, i.IsArray) for i in self.in_parameters.Properties_]
            if self.out_parameters is None:
                self.out_parameter_names = []
            else:
                self.out_parameter_names = [(i.Name, i.IsArray) for i in self.out_parameters.Properties_]

            doc = "%s (%s) => (%s)" % (
                method_name,
                ", ".join([name + ("", "[]")[is_array] for (name, is_array) in self.in_parameter_names]),
                ", ".join([name + ("", "[]")[is_array] for (name, is_array) in self.out_parameter_names])
            )
            privileges = self.qualifiers.get("Privileges", [])
            if privileges:
                doc += " | Needs: " + ", ".join(privileges)
            self.__doc__ = doc
        except pywintypes.com_error:
            handle_com_error()

    def __call__(self, *args, **kwargs):
        """Execute the call to a WMI method, returning
        a tuple (even if is of only one value) containing
        the out and return parameters.
        """
        try:
            if self.in_parameters:
                parameter_names = {}
                for name, is_array in self.in_parameter_names:
                    parameter_names[name] = is_array

                parameters = self.in_parameters

                #
                # Check positional parameters first
                #
                for n_arg in range(len(args)):
                    arg = args[n_arg]
                    parameter = parameters.Properties_[n_arg]
                    if parameter.IsArray:
                        try:
                            list(arg)
                        except TypeError:
                            raise TypeError("parameter %d must be iterable" % n_arg)
                    parameter.Value = arg

                #
                # If any keyword param supersedes a positional one,
                # it'll simply overwrite it.
                #
                for k, v in kwargs.items():
                    is_array = parameter_names.get(k)
                    if is_array is None:
                        raise AttributeError("%s is not a valid parameter for %s" % (k, self.__doc__))
                    else:
                        if is_array:
                            try:
                                list(v)
                            except TypeError:
                                raise TypeError("%s must be iterable" % k)
                    parameters.Properties_(k).Value = v

                result = self.ole_object.ExecMethod_(self.method.Name, self.in_parameters)
            else:
                result = self.ole_object.ExecMethod_(self.method.Name)

            results = []
            for name, is_array in self.out_parameter_names:
                value = result.Properties_(name).Value
                if is_array:
                    #
                    # Thanks to Jonas Bjering for bug report and patch
                    #
                    results.append(list(value or []))
                else:
                    results.append(value)
            return tuple(results)

        except pywintypes.com_error:
            handle_com_error()

    def __repr__(self):
        return "<function %s>" % self.__doc__


class _wmi_property(object):

    def __init__(self, property):
        self.property = property
        self.name = property.Name
        self.value = property.Value
        self.qualifiers = dict((q.Name, q.Value) for q in property.Qualifiers_)
        self.type = self.qualifiers.get("CIMTYPE", None)

    def set(self, value):
        self.property.Value = value

    def __repr__(self):
        return "<wmi_property: %s>" % self.name

    def __getattr__(self, attr):
        return getattr(self.property, attr)


#
# class _wmi_object
#
class _wmi_object:
    """The heart of the WMI module: wraps the objects returned by COM
    ISWbemObject interface and provide readier access to their properties
    and methods resulting in a more Pythonic interface. Not usually
    instantiated directly, rather as a result of calling a :class:`_wmi_class`
    on the parent :class:`_wmi_namespace`.

    If you get hold of a WMI-related COM object from some other
    source than this module, you can wrap it in one of these objects
    to get the benefits of the module::

      import win32com.client
      import wmi

      wmiobj = win32com.client.GetObject ("winmgmts:Win32_LogicalDisk.DeviceID='C:'")
      c_drive = wmi._wmi_object (wmiobj)
      print c_drive
    """

    def __init__(self, ole_object, instance_of=None, fields=[], property_map={}):
        try:
            _set(self, "ole_object", ole_object)
            _set(self, "id", ole_object.Path_.DisplayName.lower())
            _set(self, "_instance_of", instance_of)
            _set(self, "properties", {})
            _set(self, "methods", {})
            _set(self, "property_map", property_map)
            _set(self, "_associated_classes", None)
            _set(self, "_keys", None)

            if fields:
                for field in fields:
                    self.properties[field] = None
            else:
                for p in ole_object.Properties_:
                    self.properties[p.Name] = None

            for m in ole_object.Methods_:
                self.methods[m.Name] = None

            _set(self, "_properties", self.properties.keys())
            _set(self, "_methods", self.methods.keys())
            _set(self, "qualifiers", dict((q.Name, q.Value) for q in self.ole_object.Qualifiers_))

        except pywintypes.com_error:
            handle_com_error()

    def __lt__(self, other):
        return self.id < other.id

    def __str__(self):
        """For a call to print [object] return the OLE description
        of the properties / values of the object
        """
        try:
            return self.ole_object.GetObjectText_()
        except pywintypes.com_error:
            handle_com_error()

    def __repr__(self):
        """
        Indicate both the fact that this is a wrapped WMI object
        and the WMI object's own identifying class.
        """
        try:
            return "<%s: %s>" % (self.__class__.__name__, self.Path_.Path.encode("ascii", "backslashreplace"))
        except pywintypes.com_error:
            handle_com_error()

    def _cached_properties(self, attribute):
        if self.properties[attribute] is None:
            self.properties[attribute] = _wmi_property(self.ole_object.Properties_(attribute))
        return self.properties[attribute]

    def _cached_methods(self, attribute):
        if self.methods[attribute] is None:
            self.methods[attribute] = _wmi_method(self.ole_object, attribute)
        return self.methods[attribute]

    def __getattr__(self, attribute):
        """
        Attempt to pass attribute calls to the proxied COM object.
        If the attribute is recognised as a property, return its value;
        if it is recognised as a method, return a method wrapper which
        can then be called with parameters; otherwise pass the lookup
        on to the underlying object.
        """
        try:
            if attribute in self.properties:
                property = self._cached_properties(attribute)
                factory = self.property_map.get(attribute, self.property_map.get(property.type, lambda x: x))
                value = factory(property.value)
                #
                # If this is an association, certain of its properties
                # are actually the paths to the aspects of the association,
                # so translate them automatically into WMI objects.
                #
                if property.type.startswith("ref:"):
                    return WMI(moniker=value)
                else:
                    return value
            elif attribute in self.methods:
                return self._cached_methods(attribute)
            else:
                return getattr(self.ole_object, attribute)
        except pywintypes.com_error:
            handle_com_error()

    def __setattr__(self, attribute, value):
        """If the attribute to be set is valid for the proxied
        COM object, set that objects's parameter value; if not,
        raise an exception.
        """
        try:
            if attribute in self.properties:
                self._cached_properties(attribute).set(value)
                if self.ole_object.Path_.Path:
                    self.ole_object.Put_()
            else:
                raise AttributeError(attribute)
        except pywintypes.com_error:
            handle_com_error()

    def __eq__(self, other):
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def _getAttributeNames(self):
        """Return list of methods/properties for IPython completion"""
        attribs = [str(x) for x in self.methods.keys()]
        attribs.extend([str(x) for x in self.properties.keys()])
        return attribs

    def _get_keys(self):
        """A WMI object is uniquely defined by a set of properties
        which constitute its keys. Lazily retrieves the keys for this
        instance or class.

        :returns: list of key property names
        """
        # NB You can get the keys of an instance more directly, via
        # Path\_.Keys but this doesn't apply to classes. The technique
        # here appears to work for both.
        if self._keys is None:
            _set(self, "_keys", [])
            for property in self.ole_object.Properties_:
                for qualifier in property.Qualifiers_:
                    if qualifier.Name == "key" and qualifier.Value:
                        self._keys.append(property.Name)
        return self._keys

    keys = property(_get_keys)

    def wmi_property(self, property_name):
        """Return the cached object representing one property
        of this object
        """
        return _wmi_property(self.ole_object.Properties_(property_name))

    def put(self):
        """Push all outstanding property updates back to the
        WMI database.
        """
        self.ole_object.Put_()

    def set(self, **kwargs):
        """Set several properties of the underlying object
        at one go. This is particularly useful in combination
        with the new () method below. However, an instance
        which has been spawned in this way won't have enough
        information to write pack, so only try if the
        instance has a path.
        """
        if kwargs:
            try:
                for attribute, value in kwargs.items():
                    if attribute in self.properties:
                        self._cached_properties(attribute).set(value)
                    else:
                        raise AttributeError(attribute)
                #
                # Only try to write the attributes
                #  back if the object exists.
                #
                if self.ole_object.Path_.Path:
                    self.ole_object.Put_()
            except pywintypes.com_error:
                handle_com_error()

    def path(self):
        """Return the WMI URI to this object. Can be used to
        determine the path relative to the parent namespace::

          pp0 = wmi.WMI ().Win32_ParallelPort ()[0]
          print pp0.path ().RelPath

        ..  Do more with this
        """
        try:
            return self.ole_object.Path_
        except pywintypes.com_error:
            handle_com_error()

    def derivation(self):
        """Return a tuple representing the object derivation for
        this object, with the most specific object first::

          pp0 = wmi.WMI ().Win32_ParallelPort ()[0]
          print ' <- '.join (pp0.derivation ())
        """
        try:
            return self.ole_object.Derivation_
        except pywintypes.com_error:
            handle_com_error()

    def _cached_associated_classes(self):
        if self._associated_classes is None:
            if isinstance(self, _wmi_class):
                params = {'bSchemaOnly': True}
            else:
                params = {'bClassesOnly': True}
            try:
                associated_classes = dict(
                    (assoc.Path_.Class, _wmi_class(self._namespace, assoc)) for
                    assoc in self.ole_object.Associators_(**params)
                )
                _set(self, "_associated_classes", associated_classes)
            except pywintypes.com_error:
                handle_com_error()

        return self._associated_classes

    associated_classes = property(_cached_associated_classes)

    def associators(self, wmi_association_class="", wmi_result_class=""):
        """Return a list of objects related to this one, optionally limited
        either by association class (ie the name of the class which relates
        them) or by result class (ie the name of the class which would be
        retrieved)::

          c = wmi.WMI ()
          pp = c.Win32_ParallelPort ()[0]

          for i in pp.associators (wmi_association_class="Win32_PortResource"):
            print i

          for i in pp.associators (wmi_result_class="Win32_PnPEntity"):
            print i
        """
        try:
            return [
                _wmi_object(i) for i in \
                self.ole_object.Associators_(
                    strAssocClass=wmi_association_class,
                    strResultClass=wmi_result_class
                )
            ]
        except pywintypes.com_error:
            handle_com_error()

    def references(self, wmi_class=""):
        """Return a list of associations involving this object, optionally
        limited by the result class (the name of the association class).

        NB Associations are treated specially; although WMI only returns
        the string corresponding to the instance of each associated object,
        this module will automatically convert that to the object itself::

          c =  wmi.WMI ()
          sp = c.Win32_SerialPort ()[0]

          for i in sp.references ():
            print i

          for i in sp.references (wmi_class="Win32_SerialPortSetting"):
            print i
        """
        #
        # its .Path_.RelPath property to determine the string
        #
        try:
            return [_wmi_object(i) for i in self.ole_object.References_(strResultClass=wmi_class)]
        except pywintypes.com_error:
            handle_com_error()


#
# class _wmi_event
#
class _wmi_event(_wmi_object):
    """Slight extension of the _wmi_object class to allow
    objects which are the result of events firing to return
    extra information such as the type of event.
    """
    event_type_re = re.compile("__Instance(Creation|Modification|Deletion)Event")

    def __init__(self, event, event_info, fields=[]):
        _wmi_object.__init__(self, event, fields=fields)
        _set(self, "event_type", None)
        _set(self, "timestamp", None)
        _set(self, "previous", None)

        if event_info:
            event_type = self.event_type_re.match(event_info.Path_.Class).group(1).lower()
            _set(self, "event_type", event_type)
            if hasattr(event_info, "TIME_CREATED"):
                _set(self, "timestamp", from_1601(event_info.TIME_CREATED))
            if hasattr(event_info, "PreviousInstance"):
                _set(self, "previous", event_info.PreviousInstance)


#
# class _wmi_class
#
class _wmi_class(_wmi_object):
    """Currying class to assist in issuing queries against
     a WMI namespace. The idea is that when someone issues
     an otherwise unknown method against the WMI object, if
     it matches a known WMI class a query object will be
     returned which may then be called with one or more params
     which will form the WHERE clause::

      c = wmi.WMI ()
      c_drives = c.Win32_LogicalDisk (Name='C:')
    """

    def __init__(self, namespace, wmi_class):
        _wmi_object.__init__(self, wmi_class)
        _set(self, "_class_name", wmi_class.Path_.Class)
        if namespace:
            _set(self, "_namespace", namespace)
        else:
            class_moniker = wmi_class.Path_.DisplayName
            winmgmts, namespace_moniker, class_name = class_moniker.split(":")
            namespace = _wmi_namespace(GetObject(winmgmts + ":" + namespace_moniker), False)
            _set(self, "_namespace", namespace)

    def __getattr__(self, attribute):
        try:
            if attribute in self.properties:
                return _wmi_property(self.Properties_(attribute))
            else:
                return _wmi_object.__getattr__(self, attribute)
        except pywintypes.com_error:
            handle_com_error()

    def query(self, fields=[], **where_clause):
        """Make it slightly easier to query against the class,
         by calling the namespace's query with the class preset.
         Won't work if the class has been instantiated directly.
        """
        #
        #
        if self._namespace is None:
            raise x_wmi_no_namespace("You cannot query directly from a WMI class")

        try:
            field_list = ", ".join(fields) or "*"
            wql = "SELECT " + field_list + " FROM " + self._class_name
            if where_clause:
                wql += " WHERE " + " AND ".join(["%s = %r" % (k, str(v)) for k, v in where_clause.items()])
            return self._namespace.query(wql, self, fields)
        except pywintypes.com_error:
            handle_com_error()

    __call__ = query

    def watch_for(
            self,
            notification_type="operation",
            delay_secs=1,
            fields=[],
            **where_clause
    ):
        if self._namespace is None:
            raise x_wmi_no_namespace("You cannot watch directly from a WMI class")

        valid_notification_types = ("operation", "creation", "deletion", "modification")
        if notification_type.lower() not in valid_notification_types:
            raise x_wmi("notification_type must be one of %s" % ", ".join(valid_notification_types))

        return self._namespace.watch_for(
            notification_type=notification_type,
            wmi_class=self,
            delay_secs=delay_secs,
            fields=fields,
            **where_clause
        )

    def instances(self):
        """Return a list of instances of the WMI class
        """
        try:
            return [_wmi_object(instance, self) for instance in self.Instances_()]
        except pywintypes.com_error:
            handle_com_error()

    def new(self, **kwargs):
        """This is the equivalent to the raw-WMI SpawnInstance\_
        method. Note that there are relatively few uses for
        this, certainly fewer than you might imagine. Most
        classes which need to create a new *real* instance
        of themselves, eg Win32_Process, offer a .Create
        method. SpawnInstance\_ is generally reserved for
        instances which are passed as parameters to such
        `.Create` methods, a common example being the
        `Win32_SecurityDescriptor`, passed to `Win32_Share.Create`
        and other instances which need security.

        The example here is `Win32_ProcessStartup`, which
        controls the shown/hidden state etc. of a new
        `Win32_Process` instance::

          import win32con
          import wmi
          c = wmi.WMI ()
          startup = c.Win32_ProcessStartup.new (ShowWindow=win32con.SW_SHOWMINIMIZED)
          pid, retval = c.Win32_Process.Create (
            CommandLine="notepad.exe",
            ProcessStartupInformation=startup
          )

        ..  warning::
            previous versions of this docstring illustrated using this function
            to create a new process. This is *not* a good example of its use;
            it is better handled with something like the example above.
        """
        try:
            obj = _wmi_object(self.SpawnInstance_(), self)
            obj.set(**kwargs)
            return obj
        except pywintypes.com_error:
            handle_com_error()


#
# class _wmi_result
#
class _wmi_result:
    """Simple, data only result for targeted WMI queries which request
    data only result classes via fetch_as_classes.
    """

    def __init__(self, obj, attributes):
        if attributes:
            for attr in attributes:
                self.__dict__[attr] = obj.Properties_(attr).Value
        else:
            for p in obj.Properties_:
                attr = p.Name
                self.__dict__[attr] = obj.Properties_(attr).Value


#
# class WMI
#
class _wmi_namespace:
    """A WMI root of a computer system. The classes attribute holds a list
    of the classes on offer. This means you can explore a bit with
    things like this::

      c = wmi.WMI ()
      for i in c.classes:
        if "user" in i.lower ():
          print i
    """

    def __init__(self, namespace, find_classes):
        _set(self, "_namespace", namespace)
        #
        # wmi attribute preserved for backwards compatibility
        #
        _set(self, "wmi", namespace)

        self._classes = None
        self._classes_map = {}
        #
        # Pick up the list of classes under this namespace
        #  so that they can be queried, and used as though
        #  properties of the namespace by means of the __getattr__
        #  hook below.
        # If the namespace does not support SubclassesOf, carry on
        #  regardless
        #
        if find_classes:
            _ = self.classes

    def __repr__(self):
        return "<_wmi_namespace: %s>" % self.wmi

    def __str__(self):
        return repr(self)

    def _get_classes(self):
        if self._classes is None:
            self._classes = self.subclasses_of()
        return SelfDeprecatingDict(dict.fromkeys(self._classes))

    classes = property(_get_classes)

    def get(self, moniker):
        try:
            return _wmi_object(self.wmi.Get(moniker))
        except pywintypes.com_error:
            handle_com_error()

    def handle(self):
        """The raw OLE object representing the WMI namespace"""
        return self._namespace

    def subclasses_of(self, root="", regex=r".*"):
        try:
            SubclassesOf = self._namespace.SubclassesOf
        except AttributeError:
            return set()
        else:
            return set(
                c.Path_.Class
                for c in SubclassesOf(root)
                if re.match(regex, c.Path_.Class)
            )

    def instances(self, class_name):
        """Return a list of instances of the WMI class. This is
        (probably) equivalent to querying with no qualifiers::

          wmi.WMI ().instances ("Win32_LogicalDisk")
          # should be the same as
          wmi.WMI ().Win32_LogicalDisk ()
        """
        try:
            return [_wmi_object(obj) for obj in self._namespace.InstancesOf(class_name)]
        except pywintypes.com_error:
            handle_com_error()

    def new(self, wmi_class, **kwargs):
        """This is now implemented by a call to :meth:`_wmi_class.new`"""
        return getattr(self, wmi_class).new(**kwargs)

    new_instance_of = new

    def _raw_query(self, wql):
        """Execute a WQL query and return its raw results.  Use the flags
        recommended by Microsoft to achieve a read-only, semi-synchronous
        query where the time is taken while looping through.
        NB Backslashes need to be doubled up.
        """
        flags = wbemFlagReturnImmediately | wbemFlagForwardOnly
        wql = wql.replace("\\", "\\\\")
        try:
            return self._namespace.ExecQuery(strQuery=wql, iFlags=flags)
        except pywintypes.com_error:
            handle_com_error()

    def query(self, wql, instance_of=None, fields=[]):
        """Perform an arbitrary query against a WMI object, and return
        a list of _wmi_object representations of the results.
        """
        return [_wmi_object(obj, instance_of, fields) for obj in self._raw_query(wql)]

    def fetch_as_classes(self, wmi_classname, fields=(), **where_clause):
        """Build and execute a wql query to fetch the specified list of fields from
        the specified wmi_classname + where_clause, then return the results as
        a list of simple class instances with attributes matching field_list.

        If fields is left empty, select * and pre-load all class attributes for
        each class returned.
        """
        wql = "SELECT %s FROM %s" % (fields and ", ".join(fields) or "*", wmi_classname)
        if where_clause:
            wql += " WHERE " + " AND ".join(["%s = '%s'" % (k, v) for k, v in where_clause.items()])
        return [_wmi_result(obj, fields) for obj in self._raw_query(wql)]

    def fetch_as_lists(self, wmi_classname, fields, **where_clause):
        """Build and execute a wql query to fetch the specified list of fields from
        the specified wmi_classname + where_clause, then return the results as
        a list of lists whose values correspond to field_list.
        """
        wql = "SELECT %s FROM %s" % (", ".join(fields), wmi_classname)
        if where_clause:
            wql += " WHERE " + " AND ".join(["%s = '%s'" % (k, v) for k, v in where_clause.items()])
        results = []
        for obj in self._raw_query(wql):
            results.append([obj.Properties_(field).Value for field in fields])
        return results

    def watch_for(
            self,
            raw_wql=None,
            notification_type="operation",
            wmi_class=None,
            delay_secs=1,
            fields=[],
            **where_clause
    ):
        """Set up an event tracker on a WMI event. This function
        returns an wmi_watcher which can be called to get the
        next event::

          c = wmi.WMI ()

          raw_wql = "SELECT * FROM __InstanceCreationEvent WITHIN 2 WHERE TargetInstance ISA 'Win32_Process'"
          watcher = c.watch_for (raw_wql=raw_wql)
          while 1:
            process_created = watcher ()
            print process_created.Name

          # or

          watcher = c.watch_for (
            notification_type="Creation",
            wmi_class="Win32_Process",
            delay_secs=2,
            Name='calc.exe'
          )
          calc_created = watcher ()

        Now supports timeout on the call to watcher::

          import pythoncom
          import wmi
          c = wmi.WMI (privileges=["Security"])
          watcher1 = c.watch_for (
            notification_type="Creation",
            wmi_class="Win32_NTLogEvent",
            Type="error"
          )
          watcher2 = c.watch_for (
            notification_type="Creation",
            wmi_class="Win32_NTLogEvent",
            Type="warning"
          )

          while 1:
            try:
              error_log = watcher1 (500)
            except wmi.x_wmi_timed_out:
              pythoncom.PumpWaitingMessages ()
            else:
              print error_log

            try:
              warning_log = watcher2 (500)
            except wmi.x_wmi_timed_out:
              pythoncom.PumpWaitingMessages ()
            else:
              print warning_log
        """
        if isinstance(wmi_class, _wmi_class):
            class_name = wmi_class._class_name
        else:
            class_name = wmi_class
            wmi_class = getattr(self, class_name)
        is_extrinsic = "__ExtrinsicEvent" in wmi_class.derivation()
        if raw_wql:
            wql = raw_wql
        else:
            fields = set(['TargetInstance'] + (fields or ["*"]))
            field_list = ", ".join(fields)
            if is_extrinsic:
                if where_clause:
                    where = " WHERE " + " AND ".join(["%s = '%s'" % (k, v) for k, v in where_clause.items()])
                else:
                    where = ""
                wql = "SELECT " + field_list + " FROM " + class_name + where
            else:
                if where_clause:
                    where = " AND " + " AND ".join(["TargetInstance.%s = '%s'" % (k, v) for k, v in where_clause.items()])
                else:
                    where = ""
                wql = \
                    "SELECT %s FROM __Instance%sEvent WITHIN %d WHERE TargetInstance ISA '%s' %s" % \
                    (field_list, notification_type, delay_secs, class_name, where)

        try:
            return _wmi_watcher(
                self._namespace.ExecNotificationQuery(wql),
                is_extrinsic=is_extrinsic,
                fields=fields
            )
        except pywintypes.com_error:
            handle_com_error()

    def __getattr__(self, attribute):
        """Offer WMI classes as simple attributes. Pass through any untrapped
        unattribute to the underlying OLE object. This means that new or
        unmapped functionality is still available to the module user.
        """
        #
        # Don't try to match against known classes as was previously
        # done since the list may not have been requested
        # (find_classes=False).
        #
        try:
            return self._cached_classes(attribute)
        except pywintypes.com_error:
            return getattr(self._namespace, attribute)

    def _cached_classes(self, class_name):
        """Standard caching helper which keeps track of classes
        already retrieved by name and returns the existing object
        if found. If this is the first retrieval, store it and
        pass it back
        """
        if class_name not in self._classes_map:
            self._classes_map[class_name] = _wmi_class(self, self._namespace.Get(class_name))
        return self._classes_map[class_name]

    def _getAttributeNames(self):
        """Return list of classes for IPython completion engine"""
        return [x for x in self.classes if not x.startswith('__')]


#
# class _wmi_watcher
#
class _wmi_watcher:
    """Helper class for WMI.watch_for below (qv)"""

    _event_property_map = {
        "TargetInstance": _wmi_object,
        "PreviousInstance": _wmi_object
    }

    def __init__(self, wmi_event, is_extrinsic, fields=[]):
        self.wmi_event = wmi_event
        self.is_extrinsic = is_extrinsic
        self.fields = fields

    def __call__(self, timeout_ms=-1):
        """When called, return the instance which caused the event. Supports
         timeout in milliseconds (defaulting to infinite). If the watcher
         times out, :exc:`x_wmi_timed_out` is raised. This makes it easy to support
         watching for multiple objects.
        """
        try:
            event = self.wmi_event.NextEvent(timeout_ms)
            if self.is_extrinsic:
                return _wmi_event(event, None, self.fields)
            else:
                return _wmi_event(
                    event.Properties_("TargetInstance").Value,
                    _wmi_object(event, property_map=self._event_property_map),
                    self.fields
                )
        except pywintypes.com_error:
            handle_com_error()


PROTOCOL = "winmgmts:"


def connect(
        computer="",
        impersonation_level="",
        authentication_level="",
        authority="",
        privileges="",
        moniker="",
        wmi=None,
        namespace="",
        suffix="",
        user="",
        password="",
        find_classes=False,
        debug=False
):
    """The WMI constructor can either take a ready-made moniker or as many
    parts of one as are necessary. Eg::

      c = wmi.WMI (moniker="winmgmts:{impersonationLevel=Delegate}//remote")
      # or
      c = wmi.WMI (computer="remote", privileges=["!RemoteShutdown", "Security"])

    I daren't link to a Microsoft URL; they change so often. Try Googling for
    WMI construct moniker and see what it comes back with.

    For complete control, a named argument "wmi" can be supplied, which
    should be a SWbemServices object, which you create yourself. Eg::

      loc = win32com.client.Dispatch("WbemScripting.SWbemLocator")
      svc = loc.ConnectServer(...)
      c = wmi.WMI(wmi=svc)

    This is the only way of connecting to a remote computer with a different
    username, as the moniker syntax does not allow specification of a user
    name.

    If the `wmi` parameter is supplied, all other parameters are ignored.
    """
    global _DEBUG
    _DEBUG = debug

    try:
        try:
            if wmi:
                obj = wmi

            elif moniker:
                if not moniker.startswith(PROTOCOL):
                    moniker = PROTOCOL + moniker
                obj = GetObject(moniker)

            else:
                if user:
                    if privileges or suffix:
                        raise x_wmi_authentication("You can't specify privileges or a suffix as well as a username")
                    elif computer in (None, '', '.'):
                        raise x_wmi_authentication("You can only specify user/password for a remote connection")
                    else:
                        obj = connect_server(
                            server=computer,
                            namespace=namespace,
                            user=user,
                            password=password,
                            authority=authority,
                            impersonation_level=impersonation_level,
                            authentication_level=authentication_level
                        )

                else:
                    moniker = construct_moniker(
                        computer=computer,
                        impersonation_level=impersonation_level,
                        authentication_level=authentication_level,
                        authority=authority,
                        privileges=privileges,
                        namespace=namespace,
                        suffix=suffix
                    )
                    obj = GetObject(moniker)

            wmi_type = get_wmi_type(obj)

            if wmi_type == "namespace":
                return _wmi_namespace(obj, find_classes)
            elif wmi_type == "class":
                return _wmi_class(None, obj)
            elif wmi_type == "instance":
                return _wmi_object(obj)
            else:
                raise x_wmi("Unknown moniker type")

        except pywintypes.com_error:
            handle_com_error()

    except x_wmi_uninitialised_thread:
        raise x_wmi_uninitialised_thread("WMI returned a syntax error: you're probably running inside a thread without first calling pythoncom.CoInitialize[Ex]")


WMI = connect


def construct_moniker(
        computer=None,
        impersonation_level=None,
        authentication_level=None,
        authority=None,
        privileges=None,
        namespace=None,
        suffix=None
):
    security = []
    if impersonation_level: security.append("impersonationLevel=%s" % impersonation_level)
    if authentication_level: security.append("authenticationLevel=%s" % authentication_level)
    #
    # Use of the authority descriptor is invalid on the local machine
    #
    if authority and computer: security.append("authority=%s" % authority)
    if privileges: security.append("(%s)" % ", ".join(privileges))

    moniker = [PROTOCOL]
    if security: moniker.append("{%s}!" % ",".join(security))
    if computer: moniker.append("//%s/" % computer)
    if namespace:
        parts = re.split(r"[/\\]", namespace)
        if parts[0] != 'root':
            parts.insert(0, "root")
        moniker.append("/".join(parts))
    if suffix: moniker.append(":%s" % suffix)
    return "".join(moniker)


def get_wmi_type(obj):
    try:
        path = obj.Path_
    except AttributeError:
        return "namespace"
    else:
        if path.IsClass:
            return "class"
        else:
            return "instance"


def connect_server(
        server,
        namespace="",
        user="",
        password="",
        locale="",
        authority="",
        impersonation_level="",
        authentication_level="",
        security_flags=0x80,
        named_value_set=None
):
    """Return a remote server running WMI

    :param server: name of the server
    :param namespace: namespace to connect to - defaults to whatever's defined as default
    :param user: username to connect as, either local or domain (dom\\name or user@domain for XP)
    :param password: leave blank to use current context
    :param locale: desired locale in form MS_XXXX (eg MS_409 for Am En)
    :param authority: either "Kerberos:" or an NT domain. Not needed if included in user
    :param impersonation_level: valid WMI impersonation level
    :param security_flags: if 0, connect will wait forever; if 0x80, connect will timeout at 2 mins
    :param named_value_set: typically empty, otherwise a context-specific `SWbemNamedValueSet`

    Example::

      remote_connetion = wmi.connect_server (
        server="remote_machine", user="myname", password="mypassword"
      )
      c = wmi.WMI (wmi=remote_connection)
    """
    #
    # Thanks to Matt Mercer for example code to set
    # impersonation & authentication on ConnectServer
    #
    if impersonation_level:
        try:
            impersonation = getattr(obj._constants, "wbemImpersonationLevel%s" % impersonation_level.title())
        except AttributeError:
            raise x_wmi_authentication("No such impersonation level: %s" % impersonation_level)
    else:
        impersonation = None

    if authentication_level:
        try:
            authentication = getattr(obj._constants, "wbemAuthenticationLevel%s" % authentication_level.title())
        except AttributeError:
            raise x_wmi_authentication("No such impersonation level: %s" % impersonation_level)
    else:
        authentication = None

    server = Dispatch("WbemScripting.SWbemLocator"). \
        ConnectServer(
        server,
        namespace,
        user,
        password,
        locale,
        authority,
        security_flags,
        named_value_set
    )
    if impersonation:
        server.Security_.ImpersonationLevel = impersonation
    if authentication:
        server.Security_.AuthenticationLevel = authentication
    return server


def Registry(
        computer=None,
        impersonation_level="Impersonate",
        authentication_level="Default",
        authority=None,
        privileges=None,
        moniker=None
):
    warnings.warn("This function can be implemented using wmi.WMI (namespace='DEFAULT').StdRegProv", DeprecationWarning)
    if not moniker:
        moniker = construct_moniker(
            computer=computer,
            impersonation_level=impersonation_level,
            authentication_level=authentication_level,
            authority=authority,
            privileges=privileges,
            namespace="default",
            suffix="StdRegProv"
        )

    try:
        return _wmi_object(GetObject(moniker))

    except pywintypes.com_error:
        handle_com_error()


#
# Typical use test
#
if __name__ == '__main__':
    system = WMI()
    for my_computer in system.Win32_ComputerSystem():
        print("Disks on", my_computer.Name)
        for disk in system.Win32_LogicalDisk():
            print(disk.Caption, disk.Description, disk.ProviderName or "")



import logging

from django.core.management import BaseCommand
from django.db import connection

from isc_common.models.mat_view import Mat_view

logger = logging.getLogger(__name__)


def grop_mat_views(relname=None, schemaname=None, loger_func=lambda str: print(str), prefix='tmp'):
    def dbl_qutes_str(str):
        return f'"{str}"'

    # loger_func('start: grop_tmp_mat_views')

    if relname and schemaname:
        query = Mat_view.objects.filter(relname=relname, schemaname=schemaname)
    elif relname and not schemaname:
        query = Mat_view.objects.filter(relname=relname)
    elif not relname and schemaname:
        query = Mat_view.objects.filter(schemaname=schemaname)
    else:
        query = Mat_view.objects.filter()

    for mat_view in query:
        try:
            with connection.cursor() as cursor:
                # logger.debug(f'Creating: {m_view_name}')
                if mat_view.relname.find(f'{prefix}_') != -1:
                    loger_func(f'Droping: {dbl_qutes_str(mat_view.relname)}')
                    cursor.execute(f'DROP MATERIALIZED VIEW {dbl_qutes_str(mat_view.relname)} CASCADE;')
                    loger_func(f'Droped.')
                    # print(f'Function created')
        except Exception as ex:
            raise ex

    loger_func('Cleaned Temp matViews')

class Command(BaseCommand):

    def handle(self, *args, **options):
        grop_mat_views()



from django.urls import path

from isc_common.auth.views.auth import login, changepassword, link2group, unlinkFromgroup, copyPermission, logout
from isc_common.auth.views.get_permission import get_permission_view
from isc_common.auth.views.permission import permission_view

urlpatterns = [
    path('login', login),
    path('logout', logout),
    path('changepassword', changepassword),
    path('link2group', link2group),
    path('unlinkFromgroup', unlinkFromgroup),
    path('copyPermission', copyPermission),
    path('permission', permission_view),
    path('get_permission', get_permission_view),
]



from django.urls import path

from twits.views import chat_user_user_theme

urlpatterns = [

    path('Chat_user_user_theme/Fetch/', chat_user_user_theme.Chat_user_user_theme_Fetch),
    path('Chat_user_user_theme/Add', chat_user_user_theme.Chat_user_user_theme_Add),
    path('Chat_user_user_theme/Update', chat_user_user_theme.Chat_user_user_theme_Update),
    path('Chat_user_user_theme/Remove', chat_user_user_theme.Chat_user_user_theme_Remove),
    path('Chat_user_user_theme/Lookup/', chat_user_user_theme.Chat_user_user_theme_Lookup),
    path('Chat_user_user_theme/Info/', chat_user_user_theme.Chat_user_user_theme_Info),
    path('Chat_user_user_theme/Copy', chat_user_user_theme.Chat_user_user_theme_Copy),

]



from isc_common.http.DSResponse import DSResponseUpdate, DSResponseAdd, DSResponse, JsonResponseWithException
from isc_common.http.RPCResponse import RPCResponseConstant
from isc_common.http.response import JsonResponse
from refs.models.type_param_values import Type_param_values, Type_param_valuesManager


@JsonResponseWithException()
def Type_param_values_Fetch(request):
    return JsonResponse(
        DSResponse(
            request=request,
            data=Type_param_values.objects.
                filter().
                get_range_rows1(
                request=request,
                function=Type_param_valuesManager.getRecord
            ),
            status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Type_param_values_Add(request):
    return JsonResponse(DSResponseAdd(data=Type_param_values.objects.createFromRequest(request=request), status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Type_param_values_Update(request):
    return JsonResponse(DSResponseUpdate(data=Type_param_values.objects.updateFromRequest(request), status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Type_param_values_Remove(request):
    return JsonResponse(DSResponse(request=request, data=Type_param_values.objects.deleteFromRequest(request=request), status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Type_param_values_Lookup(request):
    return JsonResponse(DSResponse(request=request, data=Type_param_values.objects.lookupFromRequest(request=request), status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Type_param_values_Info(request):
    return JsonResponse(DSResponse(request=request, data=Type_param_values.objects.get_queryset().get_info(request=request), status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Type_param_values_Copy(request):
    return JsonResponse(DSResponse(request=request, data=Type_param_values.objects.copyFromRequest(request=request), status=RPCResponseConstant.statusSuccess).response)



import logging

from django.db.models import PositiveIntegerField

from isc_common.fields.code_field import CodeStrictField
from isc_common.models.base_ref import BaseRefHierarcy, BaseRefManager, BaseRefQuerySet

logger = logging.getLogger(__name__)


class Jasper_reportsQuerySet(BaseRefQuerySet):
    pass


class Jasper_reportsManager(BaseRefManager):

    @staticmethod
    def getRecord(record):
        res = {
            'id': record.id,
            'code': record.code,
            'name': record.name,
            'description': record.description,
            'username': record.username,
            'password': record.password,
            'host': record.host,
            'port': record.port,
            'parent': record.parent.id if record.parent else None
        }
        return res

    def get_queryset(self):
        return Jasper_reportsQuerySet(self.model, using=self._db)


class Jasper_reports(BaseRefHierarcy):
    url = CodeStrictField()
    username = CodeStrictField()
    password = CodeStrictField()
    host = CodeStrictField()
    port = PositiveIntegerField()

    objects = Jasper_reportsManager()

    def __str__(self):
        return f"ID:{self.id}, code: {self.code}, name: {self.name}, description: {self.description}"

    class Meta:
        verbose_name = 'Отчеты'



from bitfield import BitHandler, Bit

from isc_common import setAttr


def StrToNumber(s1):
    if s1 == None:
        return None

    if not isinstance(s1, str):
        raise Exception(f'{s1} is not str type.')

    s = s1.replace(',', '.')

    if not isinstance(s, str):
        raise Exception(f'{s} is not str type.')
    try:
        return int(s)
    except ValueError:
        try:
            return float(s)
        except ValueError as ex:
            return None


def StrToInt(s):
    if s == None:
        return None
    try:
        return int(s)
    except ValueError as ex:
        return None


def DelProps(value):
    if isinstance(value, dict):
        for key, _value in value.items():
            if isinstance(_value, BitHandler):
                setAttr(value, key, _value._value)
            elif isinstance(_value, Bit):
                setAttr(value, key, _value.is_set)
        return value
    else:
        value


def GetPropsInt(value):
    if isinstance(value, BitHandler):
        return value._value
    else:
        value



import logging

from django.db.models import PositiveIntegerField

from isc_common.fields.code_field import CodeField
from isc_common.models.base_ref import BaseRef, BaseRefManager, BaseRefQuerySet

logger = logging.getLogger(__name__)


class ShiftsQuerySet(BaseRefQuerySet):
    def filter(self, *args, **kwargs):
        return super().filter(*args, **kwargs)


class ShiftsManager(BaseRefManager):

    @staticmethod
    def getRecord(record):
        res = {
            'id': record.id,
            'code': record.code,
            'name': record.name,
            'description': record.description,
            'color': record.color,
        }
        return res

    def get_queryset(self):
        return ShiftsQuerySet(self.model, using=self._db)


class Shifts(BaseRef):
    color = CodeField(null=True, blank=True)
    objects = ShiftsManager()

    def __str__(self):
        return f"ID={self.id}, code={self.code}, name={self.name}, description={self.description}, color={self.color}"

    class Meta:
        verbose_name = 'Смены'



import logging

from bitfield import BitField
from django.db import transaction
from django.forms import model_to_dict

from isc_common import delAttr, setAttr
from isc_common.fields.date_time_field import DateTimeField
from isc_common.http.DSRequest import DSRequest
from isc_common.models.base_ref import BaseRefManager, BaseRefQuerySet, BaseRefHierarcy
from isc_common.number import DelProps

logger = logging.getLogger(__name__)


def get_graph(request):
    from clndr.time_line import Timeline

    request = DSRequest(request=request)
    data = dict(user_id=request.user_id)
    shiftsList = Timeline(calendar_id=request.calendar_id, logger=logger).build(data=data)
    json = shiftsList.to_json()
    return json


class CalendarsQuerySet(BaseRefQuerySet):
    def filter(self, *args, **kwargs):
        return super().filter(*args, **kwargs)

    def delete(self):
        from clndr.models.calendar_shifts_days import Calendar_shifts_days
        from clndr.models.holidays import Holidays

        with transaction.atomic():
            for item in self:
                Calendar_shifts_days.objects.filter(calendar=item).delete()
                Holidays.objects.filter(calendar=item).delete()

            return super().delete()


class CalendarsManager(BaseRefManager):

    @staticmethod
    def getRecord(record):
        res = {
            'id': record.id,
            'code': record.code,
            'name': record.name,
            'full_name': record.full_name,
            'begindate': record.begindate,
            'enddate': record.enddate,
            'description': record.description,
            'props': record.props,
            'parent_id': record.parent.id if record.parent else None,
        }
        return DelProps(res)

    def get_queryset(self):
        return CalendarsQuerySet(self.model, using=self._db)

    def copyFromRequest(self, request):
        request = DSRequest(request=request)
        data = request.get_data()
        id = data.get('id')

        calendar = Calendars.objects.get(id=id)
        calendar = calendar.copy()

        calendar = model_to_dict(calendar)

        return DelProps(calendar)


class Calendars(BaseRefHierarcy):
    begindate = DateTimeField(verbose_name='Начало', db_index=True, null=True, blank=True)
    enddate = DateTimeField(verbose_name='Конец', db_index=True, null=True, blank=True)
    props = BitField(flags=(
        ('default', 'Текущий календарь (по-умолчанию)'),
    ), default=0, db_index=True)

    objects = CalendarsManager()

    def __str__(self):
        return f"ID={self.id}, begindate={self.begindate}, enddate={self.enddate}"

    def copy(self):
        from clndr.models.calendar_shifts_days import Calendar_shifts_days
        from clndr.models.holidays import Holidays

        if self.begindate == None or self.enddate == None:
            raise Exception(f'Begindate or EndDate is null.')

        with transaction.atomic():
            calendar = model_to_dict(self)
            delAttr(calendar, 'id')
            parent = calendar.get('parent')
            delAttr(calendar, 'parent')
            setAttr(calendar, 'parent_id', parent)
            calendar = Calendars.objects.create(**calendar)

            for calendar_shifts_days in Calendar_shifts_days.objects.filter(calendar=self):
                _calendar_shifts_days = calendar_shifts_days.copy()
                _calendar_shifts_days.calendar = calendar
                _calendar_shifts_days.save()

            for holiday in Holidays.objects.filter(calendar=self):
                holiday = model_to_dict(holiday)
                delAttr(holiday, 'id')
                setAttr(holiday, 'calendar', calendar)
                daytype = holiday.get('daytype')
                delAttr(holiday, 'daytype')
                setAttr(holiday, 'daytype_id', daytype)
                Holidays.objects.create(**holiday)

            return calendar

    class Meta:
        verbose_name = 'Календари'



import logging

from isc_common.models.base_ref import BaseRefQuerySet, BaseRefManager, BaseRef

logger = logging.getLogger(__name__)


class DayQuerySet(BaseRefQuerySet):
    def filter(self, *args, **kwargs):
        return super().filter(*args, **kwargs)


class DayManager(BaseRefManager):

    @staticmethod
    def getRecord(record):
        res = {
            'id': record.id,
            'code': record.code,
            'name': record.name,
            'description': record.description,
        }
        return res

    def get_queryset(self):
        return DayQuerySet(self.model, using=self._db)


class Day(BaseRef):
    objects = DayManager()

    def __str__(self):
        return f"ID={self.id}, code={self.code}, name={self.name}, description={self.description}"

    class Meta:
        verbose_name = 'День'



import logging

from bitfield import BitField

from isc_common import delAttr
from isc_common.bit import IsBitOn
from isc_common.fields.code_field import CodeStrictField
from isc_common.http.DSRequest import DSRequest
from isc_common.models.base_ref import BaseRefHierarcy, BaseRefManager, BaseRefQuerySet

logger = logging.getLogger(__name__)


class Event_typesQuerySet(BaseRefQuerySet):
    pass


class Event_typesManager(BaseRefManager):

    @staticmethod
    def getRecord(record):
        res = {
            'id': record.id,
            'code': record.code,
            'name': record.name,
            'description': record.description,
            'parent_id': record.parent.id if record.parent else None,
            'props': record.props._value,
            'isEvent': IsBitOn(record.props._value, 0),
            'compulsory_reading': IsBitOn(record.props._value, 1),
        }
        return res

    def get_queryset(self):
        return Event_typesQuerySet(self.model, using=self._db)

    def updateFromRequest(self, request, removed=None, function=None):
        request = DSRequest(request=request)
        data = request.get_data()
        _data = data.copy()
        delAttr(_data, 'id')
        delAttr(_data, 'isEvent')
        delAttr(_data, 'isFolder')
        delAttr(_data, 'compulsory_reading')

        res = super().filter(id=data.get('id')).update(**_data)
        return data


class Event_types(BaseRefHierarcy):
    code = CodeStrictField(unique=True)
    props = BitField(flags=(
        ('isEvent', 'Является событием'),  # 1
        ('compulsory_reading', 'Обязательное прочтение'),  # 2
    ), default=0, db_index=True)

    objects = Event_typesManager()

    def __str__(self):
        return f"ID:{self.id}, code: {self.code}, name: {self.name}, description: {self.description}"

    class Meta:
        verbose_name = 'Типы событий'



# Generated by Django 3.0.2 on 2020-01-22 10:29

from django.db import migrations, models
import isc_common.fields.code_field


class Migration(migrations.Migration):

    dependencies = [
        ('reports', '0002_jasper_reports_users'),
    ]

    operations = [
        migrations.AddField(
            model_name='jasper_reports',
            name='host',
            field=isc_common.fields.code_field.CodeStrictField(),
        ),
        migrations.AddField(
            model_name='jasper_reports',
            name='password',
            field=isc_common.fields.code_field.CodeStrictField(),
        ),
        migrations.AddField(
            model_name='jasper_reports',
            name='port',
            field=models.PositiveIntegerField(default=None),
        ),
        migrations.AddField(
            model_name='jasper_reports',
            name='url',
            field=isc_common.fields.code_field.CodeStrictField(),
        ),
        migrations.AddField(
            model_name='jasper_reports',
            name='username',
            field=isc_common.fields.code_field.CodeStrictField(),
        ),
    ]



# Generated by Django 2.2.4 on 2019-08-21 20:08

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('clndr', '0018_auto_20190821_1721'),
    ]

    operations = [
        migrations.DeleteModel(
            name='Holidays_view',
        ),
    ]
